from menpo.io.landmark import PTSImporter
import numpy as np


class MeshPTSImporter(PTSImporter):
    r"""
    Implements the :meth:`_build_points` method for meshes. Here, `x` is the
    first axis.

    Parameters
    ----------
    filepath : string
        Absolute filepath of the landmarks
    """

    def __init__(self, filepath):
        super(MeshPTSImporter, self).__init__(filepath)

    def _build_points(self, xs, ys):
        """
        For meshes, `axis 0 = xs` and `axis 1 = ys`. Therefore, return the
        appropriate points array ordering.

        Parameters
        ----------
        xs : (N,) ndarray
            Row vector of `x` coordinates
        ys : (N,) ndarray
            Row vector of `y` coordinates

        Returns
        -------
        points : (N, 2) ndarray
            Array with `xs` as the first axis: `[xs; ys]`
        """
        return np.hstack([xs, ys])



from __future__ import division
import numpy as np
from scipy.linalg.blas import dgemm


def eigenvalue_decomposition(S, eps=10**-10):
    r"""

    Parameters
    ----------
    S : (N, N)  ndarray
        Covariance/Scatter matrix

    Returns
    -------
    pos_eigenvectors: (N, p) ndarray
    pos_eigenvalues: (p,) ndarray
    """
    # compute eigenvalue decomposition
    eigenvalues, eigenvectors = np.linalg.eigh(S)
    # sort eigenvalues from largest to smallest
    index = np.argsort(eigenvalues)[::-1]
    eigenvalues = eigenvalues[index]
    eigenvectors = eigenvectors[:, index]

    # set tolerance limit
    limit = np.max(np.abs(eigenvalues)) * eps

    # select positive eigenvalues
    pos_index = eigenvalues > 0.0
    pos_eigenvalues = eigenvalues[pos_index]
    pos_eigenvectors = eigenvectors[:, pos_index]
    # check they are within the expected tolerance
    index = pos_eigenvalues > limit
    pos_eigenvalues = pos_eigenvalues[index]
    pos_eigenvectors = pos_eigenvectors[:, index]

    return pos_eigenvectors, pos_eigenvalues


def principal_component_decomposition(X, whiten=False, center=True,
                                      bias=False, inplace=False):
    r"""
    Apply PCA on the data matrix X. In the case where the data matrix is very
    large, it is advisable to set `inplace=True`. However, note this this
    destructively edits the data matrix by subtracting the mean inplace.

    Parameters
    ----------
    x : (n_samples, n_features) ndarray
        Training data
    whiten : bool, optional
        Normalise the eigenvectors to have unit magnitude

        Default: `False`
    center : bool, optional
        Whether to center the data matrix. If `False`, zero will be subtracted.

        Default: `True`
    bias : bool, optional
        Whether to use a biased estimate of the number of samples. If `False`,
        subtracts `1` from the number of samples.

        Default: `False`
    inplace : bool, optional
        Whether to do the mean subtracting inplace or not. This is crucial if
        the data matrix is greater than half the available memory size.

        Default: `False`

    Returns
    -------
    eigenvectors : (n_components, n_features) ndarray
        The eigenvectors of the data matrix
    eigenvalues : (n_components,) ndarray
        The positive eigenvalues from the data matrix
    mean_vector : (n_components,) ndarray
        The mean that was subtracted from the dataset
    """
    n_samples, n_features = X.shape

    if bias:
        N = n_samples
    else:
        N = n_samples - 1.0

    if center:
        # center data
        mean_vector = np.mean(X, axis=0)
    else:
        mean_vector = np.zeros(n_features)

    # This is required if the data matrix is very large!
    if inplace:
        X -= mean_vector
    else:
        X = X - mean_vector

    if n_features < n_samples:
        # compute covariance matrix
        # S:  n_features  x  n_features
        S = dgemm(alpha=1.0, a=X.T, b=X.T, trans_b=True) / N
        # S should be perfectly symmetrical, but numerical error can creep
        # in. Enforce symmetry here to avoid creating complex
        # eigenvectors from eigendecomposition
        S = (S + S.T) / 2.0

        # perform eigenvalue decomposition
        # eigenvectors:  n_features x  n_features
        # eigenvalues:   n_features
        eigenvectors, eigenvalues = eigenvalue_decomposition(S)

        if whiten:
            # whiten eigenvectors
            eigenvectors *= np.sqrt(1.0 / eigenvalues)

    else:
        # n_features > n_samples
        # compute covariance matrix
        # S:  n_samples  x  n_samples
        S = dgemm(alpha=1.0, a=X.T, b=X.T, trans_a=True) / N
        # S should be perfectly symmetrical, but numerical error can creep
        # in. Enforce symmetry here to avoid creating complex
        # eigenvectors from eigendecomposition
        S = (S + S.T) / 2.0

        # perform eigenvalue decomposition
        # eigenvectors:  n_samples  x  n_samples
        # eigenvalues:   n_samples
        eigenvectors, eigenvalues = eigenvalue_decomposition(S)

        # compute final eigenvectors
        # eigenvectors:  n_samples  x  n_features
        if whiten:
            w = (N * eigenvalues) ** -1.0
        else:
            w = np.sqrt(1.0 / (N * eigenvalues))
        eigenvectors = w * dgemm(alpha=1.0, a=X.T, b=eigenvectors.T,
                                 trans_b=True)

    # transpose eigenvectors
    # eigenvectors:  n_samples  x  n_features
    eigenvectors = eigenvectors.T

    return eigenvectors, eigenvalues, mean_vector



from copy import deepcopy

import numpy as np

from .base import Image


class BooleanImage(Image):
    r"""
    A mask image made from binary pixels. The region of the image that is
    left exposed by the mask is referred to as the 'masked region'. The
    set of 'masked' pixels is those pixels corresponding to a True value in
    the mask.

    Parameters
    -----------
    mask_data : (M, N, ..., L) ndarray
        The binary mask data. Note that there is no channel axis - a 2D Mask
        Image is built from just a 2D numpy array of mask_data.
        Automatically coerced in to boolean values.

    copy: bool, optional
        If False, the image_data will not be copied on assignment. Note that
        if the array you provide is not boolean, there **will still be copy**.
        In general this should only be used if you know what you are doing.

        Default: `False`
    """

    def __init__(self, mask_data, copy=True):
        # Add a channel dimension. We do this little reshape trick to add
        # the axis because this maintains C-contiguous'ness
        mask_data = mask_data.reshape(mask_data.shape + (1,))
        # If we are trying not to copy, but the data we have isn't boolean,
        # then unfortunately, we forced to copy anyway!
        if mask_data.dtype != np.bool:
            # Unfortunately, even if you were trying not to copy, if you don't
            # have boolean data we have to copy!
            if not copy:
                raise Warning('The copy flag was NOT honoured. '
                              'A copy HAS been made. Please use np.bool data'
                              'to avoid this.')
            mask_data = np.require(mask_data, dtype=np.bool,
                                   requirements=['C'])
        super(BooleanImage, self).__init__(mask_data, copy=copy)

    @classmethod
    def blank(cls, shape, fill=True, round='ceil', **kwargs):
        r"""
        Returns a blank :map:`BooleanImage` of the requested shape

        Parameters
        ----------
        shape : tuple or list
            The shape of the image. Any floating point values are rounded
            according to the `round` kwarg.

        fill : True or False, optional
            The mask value to be set everywhere

            Default: True (masked region is the whole image - meaning the whole
                 image is exposed)
        round: {'ceil', 'floor', 'round'}
            Rounding function to be applied to floating point shapes.

            Default: 'ceil'

        Returns
        -------
        blank_image : :map:`BooleanImage`
            A blank mask of the requested size

        """
        if round not in ['ceil', 'round', 'floor']:
            raise ValueError('round must be either ceil, round or floor')
            # Ensure that the '+' operator means concatenate tuples
        shape = tuple(getattr(np, round)(shape).astype(np.int))
        if fill:
            mask = np.ones(shape, dtype=np.bool)
        else:
            mask = np.zeros(shape, dtype=np.bool)
        return cls(mask, copy=False)

    @property
    def mask(self):
        r"""
        Returns the pixels of the mask with no channel axis. This is what
        should be used to mask any k-dimensional image.

        :type: (M, N, ..., L), np.bool ndarray
        """
        return self.pixels[..., 0]

    @property
    def n_true(self):
        r"""
        The number of `True` values in the mask

        :type: int
        """
        return np.sum(self.pixels)

    @property
    def n_false(self):
        r"""
        The number of `False` values in the mask

        :type: int
        """
        return self.n_pixels - self.n_true

    @property
    def all_true(self):
        r"""
        True iff every element of the mask is True.

        :type: bool
        """
        return np.all(self.pixels)

    @property
    def proportion_true(self):
        r"""
        The proportion of the mask which is `True`

        :type: double
        """
        return (self.n_true * 1.0) / self.n_pixels

    @property
    def proportion_false(self):
        r"""
        The proportion of the mask which is `False`

        :type: double
        """
        return (self.n_false * 1.0) / self.n_pixels

    @property
    def true_indices(self):
        r"""
        The indices of pixels that are true.

        :type: (`n_dims`, `n_true`) ndarray
        """
        if self.all_true:
            return self.indices
        else:
            # Ignore the channel axis
            return np.vstack(np.nonzero(self.pixels[..., 0])).T

    @property
    def false_indices(self):
        r"""
        The indices of pixels that are false.

        :type: (`n_dims`, `n_false`) ndarray
        """
        # Ignore the channel axis
        return np.vstack(np.nonzero(~self.pixels[..., 0])).T

    def __str__(self):
        return ('{} {}D mask, {:.1%} '
                'of which is True'.format(self._str_shape, self.n_dims,
                                          self.proportion_true))

    def copy(self):
        r"""
        Return a new image with copies of the pixels and landmarks of this
        image.

        This is an efficient copy method. If you need to copy all the state on
        the object, consider deepcopy instead.

        Returns
        -------

        image: :map:`BooleanImage`
            A new image with the same pixels and landmarks as this one,
            just copied.

        """
        new_image = BooleanImage(self.pixels[..., 0])
        new_image.landmarks = self.landmarks
        return new_image

    def from_vector(self, vector, copy=True):
        r"""
        Takes a flattened vector and returns a new :map:`BooleanImage` formed
        by reshaping the vector to the correct dimensions. Note that this is
        rebuilding a boolean image **itself** from boolean values. The mask
        is in no way interpreted in performing the operation, in contrast to
        MaskedImage, where only the masked region is used in from_vector()
        and as_vector(). Any image landmarks are transferred in the process.

        Parameters
        ----------
        vector : (`n_pixels`,) np.bool ndarray
            A flattened vector of all the pixels of a BooleanImage.

        copy : bool, optional
            If false, no copy of the vector will be taken.

            Default: True

        Returns
        -------
        image : :map:`BooleanImage`
            New BooleanImage of same shape as this image

        Raises
        ------
        Warning : If copy=False cannot be honored.

        """
        mask = BooleanImage(vector.reshape(self.shape), copy=copy)
        mask.landmarks = self.landmarks
        return mask

    def invert_inplace(self):
        r"""
        Inverts this Boolean Image inplace.

        """
        self.pixels = ~self.pixels

    def invert(self):
        r"""
        Returns a copy of this Boolean image, which is inverted.

        Returns
        -------

        inverted : :map:`BooleanImage`
            A copy of this boolean mask, where all True values are False and
            all False values are True.

        """
        inverse = self.copy()
        inverse.invert_inplace()
        return inverse

    def bounds_true(self, boundary=0, constrain_to_bounds=True):
        r"""
        Returns the minimum to maximum indices along all dimensions that the
        mask includes which fully surround the True mask values. In the case
        of a 2D Image for instance, the min and max define two corners of a
        rectangle bounding the True pixel values.

        Parameters
        ----------
        boundary : int, optional
            A number of pixels that should be added to the extent. A
            negative value can be used to shrink the bounds in.

            Default: 0

        constrain_to_bounds: bool, optional
            If True, the bounding extent is snapped to not go beyond
            the edge of the image. If False, the bounds are left unchanged.

            Default: True

        Returns
        --------
        min_b : (D,) ndarray
            The minimum extent of the True mask region with the boundary
            along each dimension. If constrain_to_bounds was True,
            is clipped to legal image bounds.

        max_b : (D,) ndarray
            The maximum extent of the True mask region with the boundary
            along each dimension. If constrain_to_bounds was True,
            is clipped to legal image bounds.
        """
        mpi = self.true_indices
        maxes = np.max(mpi, axis=0) + boundary
        mins = np.min(mpi, axis=0) - boundary
        if constrain_to_bounds:
            maxes = self.constrain_points_to_bounds(maxes)
            mins = self.constrain_points_to_bounds(mins)
        return mins, maxes

    def bounds_false(self, boundary=0, constrain_to_bounds=True):
        r"""
        Returns the minimum to maximum indices along all dimensions that the
        mask includes which fully surround the False mask values. In the case
        of a 2D Image for instance, the min and max define two corners of a
        rectangle bounding the False pixel values.

        Parameters
        ----------
        boundary : int >= 0, optional
            A number of pixels that should be added to the extent. A
            negative value can be used to shrink the bounds in.

            Default: 0

        constrain_to_bounds: bool, optional
            If True, the bounding extent is snapped to not go beyond
            the edge of the image. If False, the bounds are left unchanged.

            Default: True

        Returns
        --------
        min_b : (D,) ndarray
            The minimum extent of the False mask region with the boundary
            along each dimension. If constrain_to_bounds was True,
            is clipped to legal image bounds.

        max_b : (D,) ndarray
            The maximum extent of the False mask region with the boundary
            along each dimension. If constrain_to_bounds was True,
            is clipped to legal image bounds.
        """
        return self.invert().bounds_true(
            boundary=boundary, constrain_to_bounds=constrain_to_bounds)

    def warp_to(self, template_mask, transform, warp_landmarks=False,
                interpolator='scipy', **kwargs):
        r"""
        Warps this BooleanImage into a different reference space.

        Parameters
        ----------
        template_mask : :class:`menpo.image.boolean.BooleanImage`
            Defines the shape of the result, and what pixels should be
            sampled.
        transform : :class:`menpo.transform.base.Transform`
            Transform **from the template space back to this image**.
            Defines, for each True pixel location on the template, which pixel
            location should be sampled from on this image.
        warp_landmarks : bool, optional
            If `True`, warped_image will have the same landmark dictionary
            as self, but with each landmark updated to the warped position.

            Default: `False`
        interpolator : 'scipy' or 'c', optional
            The interpolator that should be used to perform the warp.

            Default: 'scipy'
        kwargs : dict
            Passed through to the interpolator. See `menpo.interpolation`
            for details.

        Returns
        -------
        warped_image : type(self)
            A copy of this image, warped.
        """
        # enforce the order as 0, for this boolean data, then call super
        manually_set_order = kwargs.get('order', 0)
        if manually_set_order != 0:
            raise ValueError(
                "The order of the interpolation on a boolean image has to be "
                "0 (attempted to set {})".format(manually_set_order))
        kwargs['order'] = 0
        return Image.warp_to(self, template_mask, transform,
                             warp_landmarks=warp_landmarks,
                             interpolator=interpolator, **kwargs)

    def _build_warped_image(self, template_mask, sampled_pixel_values,
                            **kwargs):
        r"""
        Builds the warped image from the template mask and
        sampled pixel values. Overridden for BooleanImage as we can't use
        the usual from_vector_inplace method.
        """
        warped_image = BooleanImage.blank(template_mask.shape)
        # As we are a mask image, we have to implement the update a little
        # more manually than other image classes.
        warped_image.pixels[warped_image.mask] = sampled_pixel_values
        return warped_image



from menpo.landmark import *
from menpo.fit.lucaskanade.appearance import *
from menpo.transform import PiecewiseAffine, ThinPlateSplines
from menpo.transform.modeldriven import GlobalMDTransform, OrthoMDTransform
from menpo.transform import AlignmentSimilarity
from menpo.fitmultilevel.featurefunctions import sparse_hog
from menpo.model.modelinstance import PDM, OrthoPDM
from menpo.fit.gradientdescent import RegularizedLandmarkMeanShift
from menpo.fitmultilevel.clm.classifierfunctions import linear_svm_lr
from .io import import_bounding_boxes

from .base import (load_database, aam_build_benchmark, aam_fit_benchmark,
                   clm_build_benchmark, clm_fit_benchmark,
                   sdm_build_benchmark, sdm_fit_benchmark,
                   convert_fitting_results_to_ced, plot_fitting_curves)


def aam_fastest_alternating_noise(training_db_path, fitting_db_path,
                                  feature_type='igo', noise_std=0.04,
                                  verbose=False, plot=False):
    # check feature
    if not isinstance(feature_type, str):
        if not hasattr(feature_type, '__call__'):
            if feature_type is not None:
                raise ValueError("feature_type must be a string or "
                                 "function/closure or None")

    # predefined options
    db_loading_options = {'crop_proportion': 0.2,
                          'convert_to_grey': True
                          }
    training_options = {'group': 'PTS',
                        'feature_type': 'igo',
                        'transform': PiecewiseAffine,
                        'trilist': ibug_68_trimesh,
                        'normalization_diagonal': None,
                        'n_levels': 3,
                        'downscale': 2,
                        'scaled_shape_models': True,
                        'pyramid_on_features': True,
                        'max_shape_components': 25,
                        'max_appearance_components': 250,
                        'boundary': 3,
                        'interpolator': 'scipy'
                        }
    fitting_options = {'algorithm': AlternatingInverseCompositional,
                       'md_transform': OrthoMDTransform,
                       'global_transform': AlignmentSimilarity,
                       'n_shape': [3, 6, 12],
                       'n_appearance': 50,
                       'max_iters': 50,
                       'error_type': 'me_norm'
                       }
    perturb_options = {'noise_std': 0.04,
                       'rotation': False}

    # set passed parameters
    training_options['feature_type'] = feature_type
    perturb_options['noise_std'] = noise_std

    # run experiment
    training_images = load_database(training_db_path,
                                    db_loading_options=db_loading_options,
                                    verbose=verbose)
    aam = aam_build_benchmark(training_images,
                              training_options=training_options,
                              verbose=verbose)
    fitting_images = load_database(fitting_db_path,
                                   db_loading_options=db_loading_options,
                                   verbose=verbose)
    fitting_results = aam_fit_benchmark(fitting_images, aam,
                                        perturb_options=perturb_options,
                                        fitting_options=fitting_options,
                                        verbose=verbose)

    # convert results
    max_error_bin = 0.05
    bins_error_step = 0.005
    final_error_curve, initial_error_curve, error_bins = \
        convert_fitting_results_to_ced(fitting_results,
                                       max_error_bin=max_error_bin,
                                       bins_error_step=bins_error_step)

    # plot results
    if plot:
        title = "AAMs using {} and Alternating IC".format(
            training_options['feature_type'])
        y_axis = [final_error_curve, initial_error_curve]
        legend = ['Fitting', 'Initialization']
        plot_fitting_curves(error_bins, y_axis, title, new_figure=True,
                            x_limit=max_error_bin, legend=legend,
                            color_list=['r', 'b'], marker_list=['o', 'x'])
    return fitting_results, final_error_curve, initial_error_curve, error_bins


def aam_fastest_alternating_bbox(training_db_path, fitting_db_path,
                                 fitting_bboxes_path, feature_type='igo',
                                 verbose=False, plot=False):
    # check feature
    if not isinstance(feature_type, str):
        if not hasattr(feature_type, '__call__'):
            if feature_type is not None:
                raise ValueError("feature_type must be a string or "
                                 "function/closure or None")

    # predefined options
    db_loading_options = {'crop_proportion': 0.1,
                          'convert_to_grey': True
    }
    training_options = {'group': 'PTS',
                        'feature_type': 'igo',
                        'transform': PiecewiseAffine,
                        'trilist': ibug_68_trimesh,
                        'normalization_diagonal': None,
                        'n_levels': 3,
                        'downscale': 2,
                        'scaled_shape_models': True,
                        'pyramid_on_features': False,
                        'max_shape_components': 25,
                        'max_appearance_components': 250,
                        'boundary': 3,
                        'interpolator': 'scipy'
    }
    fitting_options = {'algorithm': AlternatingInverseCompositional,
                       'md_transform': OrthoMDTransform,
                       'global_transform': AlignmentSimilarity,
                       'n_shape': [3, 6, 12],
                       'n_appearance': 50,
                       'max_iters': 50,
                       'error_type': 'me_norm'
    }

    # set passed parameters
    training_options['feature_type'] = feature_type

    # run experiment
    training_images = load_database(training_db_path,
                                    db_loading_options=db_loading_options,
                                    verbose=verbose)
    aam = aam_build_benchmark(training_images,
                              training_options=training_options,
                              verbose=verbose)

    # import bounding boxes
    bboxes_list = import_bounding_boxes(fitting_bboxes_path)

    # for all fittings, we crop to 0.5
    db_loading_options['crop_proportion'] = 0.5
    fitting_images = load_database(fitting_db_path,
                                   db_loading_options=db_loading_options,
                                   bounding_boxes=bboxes_list,
                                   verbose=verbose)

    fitting_results = aam_fit_benchmark(fitting_images, aam,
                                        fitting_options=fitting_options,
                                        verbose=verbose)

    # convert results
    max_error_bin = 0.05
    bins_error_step = 0.005
    final_error_curve, initial_error_curve, error_bins = \
        convert_fitting_results_to_ced(fitting_results,
                                       max_error_bin=max_error_bin,
                                       bins_error_step=bins_error_step)

    # plot results
    if plot:
        title = "AAMs using {} and Alternating IC".format(
            training_options['feature_type'])
        y_axis = [final_error_curve, initial_error_curve]
        legend = ['Fitting', 'Initialization']
        plot_fitting_curves(error_bins, y_axis, title, new_figure=True,
                            x_limit=max_error_bin, legend=legend,
                            color_list=['r', 'b'], marker_list=['o', 'x'])
    return fitting_results, final_error_curve, initial_error_curve, error_bins


def aam_best_performance_alternating_noise(training_db_path, fitting_db_path,
                                           feature_type='igo', noise_std=0.04,
                                           verbose=False, plot=False):
    # check feature
    if not isinstance(feature_type, str):
        if not hasattr(feature_type, '__call__'):
            if feature_type is not None:
                raise ValueError("feature_type must be a string or "
                                 "function/closure or None")

    # predefined options
    db_loading_options = {'crop_proportion': 0.2,
                          'convert_to_grey': True
                          }
    training_options = {'group': 'PTS',
                        'feature_type': 'igo',
                        'transform': PiecewiseAffine,
                        'trilist': ibug_68_trimesh,
                        'normalization_diagonal': None,
                        'n_levels': 3,
                        'downscale': 1.2,
                        'scaled_shape_models': False,
                        'pyramid_on_features': True,
                        'max_shape_components': 25,
                        'max_appearance_components': 250,
                        'boundary': 3,
                        'interpolator': 'scipy'
                        }
    fitting_options = {'algorithm': AlternatingInverseCompositional,
                       'md_transform': OrthoMDTransform,
                       'global_transform': AlignmentSimilarity,
                       'n_shape': [3, 6, 12],
                       'n_appearance': 50,
                       'max_iters': 50,
                       'error_type': 'me_norm'
                       }
    perturb_options = {'noise_std': 0.04,
                       'rotation': False}

    # set passed parameters
    training_options['feature_type'] = feature_type
    perturb_options['noise_std'] = noise_std

    # run experiment
    training_images = load_database(training_db_path,
                                    db_loading_options=db_loading_options,
                                    verbose=verbose)
    aam = aam_build_benchmark(training_images,
                              training_options=training_options,
                              verbose=verbose)
    fitting_images = load_database(fitting_db_path,
                                   db_loading_options=db_loading_options,
                                   verbose=verbose)
    fitting_results = aam_fit_benchmark(fitting_images, aam,
                                        perturb_options=perturb_options,
                                        fitting_options=fitting_options,
                                        verbose=verbose)

    # convert results
    max_error_bin = 0.05
    bins_error_step = 0.005
    final_error_curve, initial_error_curve, error_bins = \
        convert_fitting_results_to_ced(fitting_results,
                                       max_error_bin=max_error_bin,
                                       bins_error_step=bins_error_step)

    # plot results
    if plot:
        title = "AAMs using {} and Alternating IC".format(
            training_options['feature_type'])
        y_axis = [final_error_curve, initial_error_curve]
        legend = ['Fitting', 'Initialization']
        plot_fitting_curves(error_bins, y_axis, title, new_figure=True,
                            x_limit=max_error_bin, legend=legend,
                            color_list=['r', 'b'], marker_list=['o', 'x'])
    return fitting_results, final_error_curve, initial_error_curve, error_bins


def aam_best_performance_alternating_bbox(training_db_path, fitting_db_path,
                                          fitting_bboxes_path,
                                          feature_type='igo', verbose=False,
                                          plot=False):
    # check feature
    if not isinstance(feature_type, str):
        if not hasattr(feature_type, '__call__'):
            if feature_type is not None:
                raise ValueError("feature_type must be a string or "
                                 "function/closure or None")

    # predefined options
    db_loading_options = {'crop_proportion': 0.5,
                          'convert_to_grey': True
    }
    training_options = {'group': 'PTS',
                        'feature_type': 'igo',
                        'transform': PiecewiseAffine,
                        'trilist': ibug_68_trimesh,
                        'normalization_diagonal': 200,
                        'n_levels': 3,
                        'downscale': 2,
                        'scaled_shape_models': True,
                        'pyramid_on_features': True,
                        'max_shape_components': 25,
                        'max_appearance_components': 100,
                        'boundary': 3,
                        'interpolator': 'scipy'
    }
    fitting_options = {'algorithm': AlternatingInverseCompositional,
                       'md_transform': OrthoMDTransform,
                       'global_transform': AlignmentSimilarity,
                       'n_shape': [3, 6, 12],
                       'n_appearance': 50,
                       'max_iters': 50,
                       'error_type': 'me_norm'
    }

    # set passed parameters
    training_options['feature_type'] = feature_type

    # run experiment
    training_images = load_database(training_db_path,
                                    db_loading_options=db_loading_options,
                                    verbose=verbose)
    aam = aam_build_benchmark(training_images,
                              training_options=training_options,
                              verbose=verbose)

    # import bounding boxes
    bboxes_list = import_bounding_boxes(fitting_bboxes_path)

    # for all fittings, we crop to 0.5
    db_loading_options['crop_proportion'] = 0.5
    fitting_images = load_database(fitting_db_path,
                                   db_loading_options=db_loading_options,
                                   bounding_boxes=bboxes_list,
                                   verbose=verbose)

    fitting_results = aam_fit_benchmark(fitting_images, aam,
                                        fitting_options=fitting_options,
                                        verbose=verbose)

    # convert results
    max_error_bin = 0.05
    bins_error_step = 0.005
    final_error_curve, initial_error_curve, error_bins = \
        convert_fitting_results_to_ced(fitting_results,
                                       max_error_bin=max_error_bin,
                                       bins_error_step=bins_error_step)

    # plot results
    if plot:
        title = "AAMs using {} and Alternating IC".format(
            training_options['feature_type'])
        y_axis = [final_error_curve, initial_error_curve]
        legend = ['Fitting', 'Initialization']
        plot_fitting_curves(error_bins, y_axis, title, new_figure=True,
                            x_limit=max_error_bin, legend=legend,
                            color_list=['r', 'b'], marker_list=['o', 'x'])
    return fitting_results, final_error_curve, initial_error_curve, error_bins


def clm_basic_noise(training_db_path,  fitting_db_path,
                    feature_type=sparse_hog, classifier_type=linear_svm_lr,
                    noise_std=0.04, verbose=False, plot=False):
    # check feature
    if not isinstance(feature_type, str):
        if not hasattr(feature_type, '__call__'):
            if feature_type is not None:
                raise ValueError("feature_type must be a string or "
                                 "function/closure or None")

    # predefined options
    db_loading_options = {'crop_proportion': 0.4,
                          'convert_to_grey': True
                          }
    training_options = {'group': 'PTS',
                        'classifier_type': linear_svm_lr,
                        'patch_shape': (5, 5),
                        'feature_type': sparse_hog,
                        'normalization_diagonal': None,
                        'n_levels': 3,
                        'downscale': 1.1,
                        'scaled_shape_models': True,
                        'pyramid_on_features': False,
                        'max_shape_components': None,
                        'boundary': 3,
                        'interpolator': 'scipy'
                        }
    fitting_options = {'algorithm': RegularizedLandmarkMeanShift,
                       'pdm_transform': OrthoPDM,
                       'global_transform': AlignmentSimilarity,
                       'n_shape': [3, 6, 12],
                       'max_iters': 50,
                       'error_type': 'me_norm'
                       }
    perturb_options = {'noise_std': 0.01,
                       'rotation': False}

    # set passed parameters
    training_options['feature_type'] = feature_type
    training_options['classifier_type'] = classifier_type
    perturb_options['noise_std'] = noise_std

    # run experiment
    training_images = load_database(training_db_path,
                                    db_loading_options=db_loading_options,
                                    verbose=verbose)
    clm = clm_build_benchmark(training_images,
                              training_options=training_options,
                              verbose=verbose)
    fitting_images = load_database(fitting_db_path,
                                   db_loading_options=db_loading_options,
                                   verbose=verbose)
    fitting_results = clm_fit_benchmark(fitting_images, clm,
                                        perturb_options=perturb_options,
                                        fitting_options=fitting_options,
                                        verbose=verbose)

    # convert results
    max_error_bin = 0.05
    bins_error_step = 0.005
    final_error_curve, initial_error_curve, error_bins = \
        convert_fitting_results_to_ced(fitting_results,
                                       max_error_bin=max_error_bin,
                                       bins_error_step=bins_error_step)

    # plot results
    if plot:
        title = "CLMs with {} and {} classifier using RLMS".format(
            training_options['feature_type'],
            training_options['classifier_type'])
        y_axis = [final_error_curve, initial_error_curve]
        legend = ['Fitting', 'Initialization']
        plot_fitting_curves(error_bins, y_axis, title, new_figure=True,
                            x_limit=max_error_bin, legend=legend,
                            color_list=['r', 'b'], marker_list=['o', 'x'])
    return fitting_results, final_error_curve, initial_error_curve, error_bins


def clm_basic_bbox(training_db_path,  fitting_db_path, fitting_bboxes_path,
                   feature_type=sparse_hog, classifier_type=linear_svm_lr,
                   verbose=False, plot=False):
    # check feature
    if not isinstance(feature_type, str):
        if not hasattr(feature_type, '__call__'):
            if feature_type is not None:
                raise ValueError("feature_type must be a string or "
                                 "function/closure or None")

    # predefined options
    db_loading_options = {'crop_proportion': 0.5,
                          'convert_to_grey': True
    }
    training_options = {'group': 'PTS',
                        'classifier_type': linear_svm_lr,
                        'patch_shape': (5, 5),
                        'feature_type': sparse_hog,
                        'normalization_diagonal': None,
                        'n_levels': 3,
                        'downscale': 1.1,
                        'scaled_shape_models': True,
                        'pyramid_on_features': False,
                        'max_shape_components': None,
                        'boundary': 3,
                        'interpolator': 'scipy'
    }
    fitting_options = {'algorithm': RegularizedLandmarkMeanShift,
                       'pdm_transform': OrthoPDM,
                       'global_transform': AlignmentSimilarity,
                       'n_shape': [3, 6, 12],
                       'max_iters': 50,
                       'error_type': 'me_norm'
    }

    # set passed parameters
    training_options['feature_type'] = feature_type
    training_options['classifier_type'] = classifier_type

    # run experiment
    training_images = load_database(training_db_path,
                                    db_loading_options=db_loading_options,
                                    verbose=verbose)

    clm = clm_build_benchmark(training_images,
                              training_options=training_options,
                              verbose=verbose)

    # import bounding boxes
    bboxes_list = import_bounding_boxes(fitting_bboxes_path)

    # for all fittings, we crop to 0.5
    db_loading_options['crop_proportion'] = 0.5
    fitting_images = load_database(fitting_db_path,
                                   db_loading_options=db_loading_options,
                                   bounding_boxes=bboxes_list,
                                   verbose=verbose)

    fitting_results = clm_fit_benchmark(fitting_images, clm,
                                        fitting_options=fitting_options,
                                        verbose=verbose)

    # convert results
    max_error_bin = 0.05
    bins_error_step = 0.005
    final_error_curve, initial_error_curve, error_bins = \
        convert_fitting_results_to_ced(fitting_results,
                                       max_error_bin=max_error_bin,
                                       bins_error_step=bins_error_step)

    # plot results
    if plot:
        title = "CLMs with {} and {} classifier using RLMS".format(
            training_options['feature_type'],
            training_options['classifier_type'])
        y_axis = [final_error_curve, initial_error_curve]
        legend = ['Fitting', 'Initialization']
        plot_fitting_curves(error_bins, y_axis, title, new_figure=True,
                            x_limit=max_error_bin, legend=legend,
                            color_list=['r', 'b'], marker_list=['o', 'x'])
    return fitting_results, final_error_curve, initial_error_curve, error_bins


def sdm_fastest_bbox(training_db_path, fitting_db_path,
                                 fitting_bboxes_path, feature_type=None,
                                 verbose=False, plot=False):
    # check feature
    if not isinstance(feature_type, str):
        if not hasattr(feature_type, '__call__'):
            if feature_type is not None:
                raise ValueError("feature_type must be a string or "
                                 "function/closure or None")

    # predefined options
    db_loading_options = {'crop_proportion': 0.8,
                          'convert_to_grey': True
    }
    training_options = {'group': 'PTS',
                        'normalization_diagonal': 200,
                        'n_levels': 4,
                        'downscale': 1.01,
                        'noise_std': 0.08,
                        'patch_shape': (16, 16),
                        'n_perturbations': 15,
                        'pyramid_on_features': False,
                        'interpolator': 'scipy'
    }
    fitting_options = {
                       'error_type': 'me_norm'
    }

    # run experiment
    training_images = load_database(training_db_path,
                                    db_loading_options=db_loading_options,
                                    verbose=verbose)
    sdm = sdm_build_benchmark(training_images,
                              training_options=training_options,
                              verbose=verbose)

    # import bounding boxes
    bboxes_list = import_bounding_boxes(fitting_bboxes_path)
    # predefined options

    # for all fittings, we crop to 0.5
    db_loading_options['crop_proportion'] = 0.5
    fitting_images = load_database(fitting_db_path,
                                   db_loading_options=db_loading_options,
                                   bounding_boxes=bboxes_list,
                                   verbose=verbose)

    fitting_results = sdm_fit_benchmark(fitting_images, sdm,
                                        fitting_options=fitting_options,
                                        verbose=verbose)

    # convert results
    max_error_bin = 0.05
    bins_error_step = 0.005
    final_error_curve, initial_error_curve, error_bins = \
        convert_fitting_results_to_ced(fitting_results,
                                       max_error_bin=max_error_bin,
                                       bins_error_step=bins_error_step)

    # plot results
    if plot:
        title = "SDMs using default (sparse hogs)".format(
            training_options['feature_type'])
        y_axis = [final_error_curve, initial_error_curve]
        legend = ['Fitting', 'Initialization']
        plot_fitting_curves(error_bins, y_axis, title, new_figure=True,
                            x_limit=max_error_bin, legend=legend,
                            color_list=['r', 'b'], marker_list=['o', 'x'])
    return fitting_results, final_error_curve, initial_error_curve, error_bins


def aam_params_combinations_noise(training_db_path, fitting_db_path,
                                  n_experiments=1, feature_type=None,
                                  scaled_shape_models=None,
                                  pyramid_on_features=None, n_shape=None,
                                  n_appearance=None, noise_std=None,
                                  rotation=None, verbose=False, plot=False):

    # parse input
    if feature_type is None:
        feature_type = ['igo'] * n_experiments
    elif len(feature_type) is not n_experiments:
        raise ValueError("feature_type has wrong length")
    if scaled_shape_models is None:
        scaled_shape_models = [True] * n_experiments
    elif len(scaled_shape_models) is not n_experiments:
        raise ValueError("scaled_shape_models has wrong length")
    if pyramid_on_features is None:
        pyramid_on_features = [True] * n_experiments
    elif len(pyramid_on_features) is not n_experiments:
        raise ValueError("pyramid_on_features has wrong length")
    if n_shape is None:
        n_shape = [[3, 6, 12]] * n_experiments
    elif len(n_shape) is not n_experiments:
        raise ValueError("n_shape has wrong length")
    if n_appearance is None:
        n_appearance = [50] * n_experiments
    elif len(n_appearance) is not n_experiments:
        raise ValueError("n_appearance has wrong length")
    if noise_std is None:
        noise_std = [0.04] * n_experiments
    elif len(noise_std) is not n_experiments:
        raise ValueError("noise_std has wrong length")
    if rotation is None:
        rotation = [False] * n_experiments
    elif len(rotation) is not n_experiments:
        raise ValueError("rotation has wrong length")

    # load images
    db_loading_options = {'crop_proportion': 0.1,
                          'convert_to_grey': True
                          }
    training_images = load_database(training_db_path,
                                    db_loading_options=db_loading_options,
                                    verbose=verbose)
    fitting_images = load_database(fitting_db_path,
                                   db_loading_options=db_loading_options,
                                   verbose=verbose)

    # run experiments
    max_error_bin = 0.05
    bins_error_step = 0.005
    curves_to_plot = []
    all_fitting_results = []
    for i in range(n_experiments):
        if verbose:
            print("\nEXPERIMENT {}/{}:".format(i + 1, n_experiments))
            print("- feature_type: {}\n- scaled_shape_models: {}\n"
                  "- pyramid_on_features: {}\n- n_shape: {}\n"
                  "- n_appearance: {}\n- noise_std: {}\n"
                  "- rotation: {}".format(
                  feature_type[i], scaled_shape_models[i],
                  pyramid_on_features[i], n_shape[i], n_appearance[i],
                  noise_std[i], rotation[i]))

        # predefined option dictionaries
        training_options = {'group': 'PTS',
                            'feature_type': 'igo',
                            'transform': PiecewiseAffine,
                            'trilist': ibug_68_trimesh,
                            'normalization_diagonal': None,
                            'n_levels': 3,
                            'downscale': 1.1,
                            'scaled_shape_models': True,
                            'pyramid_on_features': True,
                            'max_shape_components': 25,
                            'max_appearance_components': 250,
                            'boundary': 3,
                            'interpolator': 'scipy'
                            }
        fitting_options = {'algorithm': AlternatingInverseCompositional,
                           'md_transform': OrthoMDTransform,
                           'global_transform': AlignmentSimilarity,
                           'n_shape': [3, 6, 12],
                           'n_appearance': 50,
                           'max_iters': 50,
                           'error_type': 'me_norm'
                           }
        pertrub_options = {'noise_std': 0.04,
                           'rotation': False}

        # training
        training_options['feature_type'] = feature_type[i]
        training_options['scaled_shape_models'] = scaled_shape_models[i]
        training_options['pyramid_on_features'] = pyramid_on_features[i]
        aam = aam_build_benchmark(training_images,
                                  training_options=training_options,
                                  verbose=verbose)

        # fitting
        fitting_options['n_shape'] = n_shape[i]
        fitting_options['n_appearance'] = n_appearance[i]
        pertrub_options['noise_std'] = noise_std[i]
        pertrub_options['rotation'] = rotation[i]
        fitting_results = aam_fit_benchmark(fitting_images, aam,
                                            perturb_options=pertrub_options,
                                            fitting_options=fitting_options,
                                            verbose=verbose)
        all_fitting_results.append(fitting_results)

        # convert results
        final_error_curve, initial_error_curve, error_bins = \
            convert_fitting_results_to_ced(fitting_results,
                                           max_error_bin=max_error_bin,
                                           bins_error_step=bins_error_step)
        curves_to_plot.append(final_error_curve)
        if i == n_experiments - 1:
            curves_to_plot.append(initial_error_curve)

    # plot results
    if plot:
        title = "AAMs using Alternating IC"
        color_list = ['r', 'b', 'g', 'y', 'c'] * n_experiments
        marker_list = ['o', 'x', 'v', 'd'] * n_experiments
        plot_fitting_curves(error_bins, curves_to_plot, title, new_figure=True,
                            x_limit=max_error_bin,  color_list=color_list,
                            marker_list=marker_list)
    return all_fitting_results


def clm_params_combinations_noise(training_db_path, fitting_db_path,
                                  n_experiments=1, classifier_type=None,
                                  patch_shape=None, feature_type=None,
                                  scaled_shape_models=None,
                                  pyramid_on_features=None, n_shape=None,
                                  noise_std=None, rotation=None, verbose=False,
                                  plot=False):

    # parse input
    if classifier_type is None:
        classifier_type = [linear_svm_lr] * n_experiments
    elif len(classifier_type) is not n_experiments:
        raise ValueError("classifier_type has wrong length")
    if patch_shape is None:
        patch_shape = [(5, 5)] * n_experiments
    elif len(patch_shape) is not n_experiments:
        raise ValueError("patch_shape has wrong length")
    if feature_type is None:
        feature_type = ['igo'] * n_experiments
    elif len(feature_type) is not n_experiments:
        raise ValueError("feature_type has wrong length")
    if scaled_shape_models is None:
        scaled_shape_models = [True] * n_experiments
    elif len(scaled_shape_models) is not n_experiments:
        raise ValueError("scaled_shape_models has wrong length")
    if pyramid_on_features is None:
        pyramid_on_features = [True] * n_experiments
    elif len(pyramid_on_features) is not n_experiments:
        raise ValueError("pyramid_on_features has wrong length")
    if n_shape is None:
        n_shape = [[3, 6, 12]] * n_experiments
    elif len(n_shape) is not n_experiments:
        raise ValueError("n_shape has wrong length")
    if noise_std is None:
        noise_std = [0.04] * n_experiments
    elif len(noise_std) is not n_experiments:
        raise ValueError("noise_std has wrong length")
    if rotation is None:
        rotation = [False] * n_experiments
    elif len(rotation) is not n_experiments:
        raise ValueError("rotation has wrong length")

    # load images
    db_loading_options = {'crop_proportion': 0.4,
                          'convert_to_grey': True
                          }
    training_images = load_database(training_db_path,
                                    db_loading_options=db_loading_options,
                                    verbose=verbose)
    fitting_images = load_database(fitting_db_path,
                                   db_loading_options=db_loading_options,
                                   verbose=verbose)

    # run experiments
    max_error_bin = 0.05
    bins_error_step = 0.005
    curves_to_plot = []
    all_fitting_results = []
    for i in range(n_experiments):
        if verbose:
            print("\nEXPERIMENT {}/{}:".format(i + 1, n_experiments))
            print("- classifier_type: {}\n- patch_shape: {}\n"
                  "- feature_type: {}\n- scaled_shape_models: {}\n"
                  "- pyramid_on_features: {}\n- n_shape: {}\n"
                  "- noise_std: {}\n- rotation: {}".format(
                  classifier_type[i], patch_shape[i], feature_type[i],
                  scaled_shape_models[i], pyramid_on_features[i], n_shape[i],
                  noise_std[i], rotation[i]))

        # predefined option dictionaries
        training_options = {'group': 'PTS',
                            'classifier_type': linear_svm_lr,
                            'patch_shape': (5, 5),
                            'feature_type': sparse_hog,
                            'normalization_diagonal': None,
                            'n_levels': 3,
                            'downscale': 1.1,
                            'scaled_shape_models': False,
                            'pyramid_on_features': True,
                            'max_shape_components': None,
                            'boundary': 3,
                            'interpolator': 'scipy'
                            }
        fitting_options = {'algorithm': RegularizedLandmarkMeanShift,
                           'pdm_transform': OrthoPDM,
                           'global_transform': AlignmentSimilarity,
                           'n_shape': [3, 6, 12],
                           'max_iters': 50,
                           'error_type': 'me_norm'
                           }
        perturb_options = {'noise_std': 0.01,
                           'rotation': False}

        # training
        training_options['classifier_type'] = classifier_type[i]
        training_options['patch_shape'] = patch_shape[i]
        training_options['feature_type'] = feature_type[i]
        training_options['scaled_shape_models'] = scaled_shape_models[i]
        training_options['pyramid_on_features'] = pyramid_on_features[i]
        clm = clm_build_benchmark(training_images,
                                  training_options=training_options,
                                  verbose=verbose)

        # fitting
        fitting_options['n_shape'] = n_shape[i]
        perturb_options['noise_std'] = noise_std[i]
        perturb_options['rotation'] = rotation[i]
        fitting_results = clm_fit_benchmark(fitting_images, clm,
                                            perturb_options=perturb_options,
                                            fitting_options=fitting_options,
                                            verbose=verbose)
        all_fitting_results.append(fitting_results)

        # convert results
        final_error_curve, initial_error_curve, error_bins = \
            convert_fitting_results_to_ced(fitting_results,
                                           max_error_bin=max_error_bin,
                                           bins_error_step=bins_error_step)
        curves_to_plot.append(final_error_curve)
        if i == n_experiments - 1:
            curves_to_plot.append(initial_error_curve)

    # plot results
    if plot:
        title = "CLMs using RLMS"
        color_list = ['r', 'b', 'g', 'y', 'c'] * n_experiments
        marker_list = ['o', 'x', 'v', 'd'] * n_experiments
        plot_fitting_curves(error_bins, curves_to_plot, title, new_figure=True,
                            x_limit=max_error_bin,  color_list=color_list,
                            marker_list=marker_list)
    return all_fitting_results



import numpy as np
from scipy.linalg.blas import dgemm
from menpo.math import principal_component_decomposition
from menpo.model.base import MeanInstanceLinearModel


class PCAModel(MeanInstanceLinearModel):
    """A :map:`MeanInstanceLinearModel` where components Principal Components.


    Principal Component Analysis (PCA) by eigenvalue decomposition of the
    data's scatter matrix. For details of the implementation of PCA, see
    :map:`principal_component_decomposition`.

    Parameters
    ----------
    samples : list of :map:`Vectorizable`
        List of samples to build the model from.
    center : bool, optional
        When True (True by default) PCA is performed after mean centering the
        data. If False the data is assumed to be centred, and the mean will
        be 0.
    bias: bool, optional
        When True (False by default) a biased estimator of the covariance
        matrix is used. See notes.

    ..notes:

    True bias mean that we calculate the covariance as

    :math:`\frac{1}{N} \sum_i^N \mathbf{x}_i \mathbf{x}_i^T`

    instead of default

    :math:`\frac{1}{N-1} \sum_i^N \mathbf{x}_i \mathbf{x}_i^T`

    """
    def __init__(self, samples, center=True, bias=False):
        # build data matrix
        n_samples = len(samples)
        n_features = samples[0].n_parameters
        data = np.zeros((n_samples, n_features))
        for i, sample in enumerate(samples):
            data[i] = sample.as_vector()

        # compute pca
        eigenvectors, eigenvalues, mean_vector = \
            principal_component_decomposition(data, whiten=False,
                                              center=center, bias=bias)

        super(PCAModel, self).__init__(eigenvectors, mean_vector, samples[0])
        self.centered = center
        self.biased = bias
        self._eigenvalues = eigenvalues
        self._n_components = self.n_components
        self._trimmed_eigenvalues = None

    @property
    def n_active_components(self):
        r"""
        The number of components currently in use on this model.
        """
        return self._n_components

    @n_active_components.setter
    def n_active_components(self, value):
        err_str = ("Tried setting n_active_components to {} - "
                   "value needs to be a float "
                   "0.0 < n_components < self._total_kept_variance_ratio "
                   "({}) or an integer 1 < n_components < "
                   "self.n_components ({})".format(
                   value, self._total_variance_ratio, self.n_components))

        # check value
        if isinstance(value, float):
            if 0.0 < value <= self._total_variance_ratio:
                # value needed to capture desired variance
                value = np.sum(
                    [r < value
                     for r in self._total_eigenvalues_cumulative_ratio]) + 1
            else:
                # variance must be bigger than 0.0
                raise ValueError(err_str)
        if isinstance(value, int):
            if value < 1:
                # at least 1 value must be kept
                raise ValueError(err_str)
            elif value >= self.n_components:
                if self.n_active_components < self.n_components:
                    # if the number of available components is smaller than
                    # the total number of components set value to the later
                    value = self.n_components
                else:
                    # if the previous is false and value bigger than the
                    # total number of components, do nothing
                    return

        if 0 < value <= self.n_components:
            self._n_components = value
        else:
            raise ValueError(err_str)

    @MeanInstanceLinearModel.components.getter
    def components(self):
        r"""
        Returns the active components of the model.

        type: (n_active_components, n_features) ndarray
        """
        return self._components[:self.n_active_components, :]

    @property
    def whitened_components(self):
        r"""
        Returns the active components of the model whitened.

        type: (n_active_components, n_features) ndarray
        """
        return self.components / (
            np.sqrt(self.eigenvalues + self.noise_variance)[:, None])

    @property
    def original_variance(self):
        r"""
        Returns the total amount of variance captured by the original model,
        i.e. the amount of variance present on the original samples.

        type: float
        """
        original_variance = self._eigenvalues.sum()
        if self._trimmed_eigenvalues is not None:
            original_variance += self._trimmed_eigenvalues.sum()
        return original_variance

    @property
    def variance(self):
        r"""
        Returns the total amount of variance retained by the active
        components.

        type: float
        """
        return self.eigenvalues.sum()

    @property
    def _total_variance(self):
        r"""
        Returns the total amount of variance retained by all components
        (active and inactive). Useful when the model has been trimmed.

        type: float
        """
        return self._eigenvalues.sum()

    @property
    def variance_ratio(self):
        r"""
        Returns the ratio between the amount of variance retained by the
        active components and the total amount of variance present on the
        original samples.

        type: float
        """
        return self.variance / self.original_variance

    @property
    def _total_variance_ratio(self):
        r"""
        Returns the ratio between the total amount of variance retained by
        all components (active and inactive) and the total amount of variance
        present on the original samples.

        type: float
        """
        return self._total_variance / self.original_variance

    @property
    def eigenvalues(self):
        r"""
        Returns the eigenvalues associated to the active components of the
        model, i.e. the amount of variance captured by each active component.

        type: (n_active_components,) ndarray
        """
        return self._eigenvalues[:self.n_active_components]

    @property
    def eigenvalues_ratio(self):
        r"""
        Returns the ratio between the variance captured by each active
        component and the total amount of variance present on the original
        samples.

        type: (n_active_components,) ndarray
        """
        return self.eigenvalues / self.original_variance

    @property
    def _total_eigenvalues_ratio(self):
        r"""
        Returns the ratio between the variance captured by each active
        component and the total amount of variance present on the original
        samples.

        type: (n_active_components,) ndarray
        """
        return self._eigenvalues / self.original_variance

    @property
    def eigenvalues_cumulative_ratio(self):
        r"""
        Returns the cumulative ratio between the variance captured by the
        active components and the total amount of variance present on the
        original samples.

        type: (n_active_components,) ndarray
        """
        cumulative_ratio = []
        previous_ratio = 0
        for ratio in self.eigenvalues_ratio:
            new_ratio = previous_ratio + ratio
            cumulative_ratio.append(new_ratio)
            previous_ratio = new_ratio
        return cumulative_ratio

    @property
    def _total_eigenvalues_cumulative_ratio(self):
        r"""
        Returns the cumulative ratio between the variance captured by the
        active components and the total amount of variance present on the
        original samples.

        type: (n_active_components,) ndarray
        """
        total_cumulative_ratio = []
        previous_ratio = 0
        for ratio in self._total_eigenvalues_ratio:
            new_ratio = previous_ratio + ratio
            total_cumulative_ratio.append(new_ratio)
            previous_ratio = new_ratio
        return total_cumulative_ratio

    @property
    def noise_variance(self):
        r"""
        Returns the average variance captured by the inactive components,
        i.e. the sample noise assumed in a PPCA formulation.

        If all components are active, noise variance is equal to 0.0

        type: float
        """
        if self.n_active_components == self.n_components:
            noise_variance = 0.0
            if self._trimmed_eigenvalues is not None:
                noise_variance += self._trimmed_eigenvalues.mean()
        else:
            if self._trimmed_eigenvalues is not None:
                noise_variance = np.hstack(
                    (self._eigenvalues[self.n_active_components:],
                     self._trimmed_eigenvalues)).mean()
            else:
                noise_variance = (
                    self._eigenvalues[self.n_active_components:].mean())
        return noise_variance

    @property
    def noise_variance_ratio(self):
        r"""
        Returns the ratio between the noise variance and the total amount of
        variance present on the original samples.

        type: float
        """
        return self.noise_variance / self.original_variance

    @property
    def inverse_noise_variance(self):
        r"""
        Returns the inverse of the noise variance.

        type: float
        """
        noise_variance = self.noise_variance
        if noise_variance == 0:
            raise ValueError("noise variance is nil - cannot take the "
                             "inverse")
        return 1.0 / noise_variance

    def component_vector(self, index, with_mean=True, scale=1.0):
        r"""
        A particular component of the model, in vectorized form.

        Parameters
        ----------
        index : int
            The component that is to be returned

        with_mean: boolean (optional)
            If True, the component will be blended with the mean vector
            before being returned. If not, the component is returned on it's
            own.

            Default: True
        scale : float
            A scale factor that should be applied to the component. Only
            valid in the case where with_mean is True. The scale is applied
            in units of standard deviations (so a scale of 1.0
            with_mean visualizes the mean plus 1 std. dev of the component
            in question).

        :type: (n_features,) ndarray
        """
        if with_mean:
            # on PCA, scale is in units of std. deviations...
            scaled_eigval = scale * np.sqrt(self.eigenvalues[index])
            return (scaled_eigval * self.components[index]) + self.mean_vector
        else:
            return self.components[index]

    def instance_vectors(self, weights):
        """
        Creates new vectorized instances of the model using the first
        components in a particular weighting.

        Parameters
        ----------
        weights : (n_vectors, n_weights) ndarray or list of lists
            The weightings for the first n_weights components that
            should be used per instance that is to be produced

            `weights[i, j]` is the linear contribution of the j'th
            principal component to the i'th instance vector produced. Note
            that if n_weights < n_components, only the first n_weight
            components are used in the reconstruction (i.e. unspecified
            weights are implicitly 0)

        Raises
        ------
        ValueError: If n_weights > n_components

        Returns
        -------
        vectors : (n_vectors, n_features) ndarray
            The instance vectors for the weighting provided.
        """
        weights = np.asarray(weights)  # if eg a list is provided
        n_instances, n_weights = weights.shape
        if n_weights > self.n_active_components:
            raise ValueError(
                "Number of weightings cannot be greater than {}".format(
                    self.n_active_components))
        else:
            full_weights = np.zeros((n_instances, self.n_active_components))
            full_weights[..., :n_weights] = weights
            weights = full_weights
        return self._instance_vectors_for_full_weights(weights)

    def trim_components(self, n_components=None):
        r"""
        Permanently trims the components down to a certain amount. The
        number of active components will be automatically reset to this
        particular value.

        This will reduce `self.n_components` down to `n_components`
        (if None `self.n_active_components` will be used), freeing
        up memory in the process.

        Once the model is trimmed, the trimmed components cannot be recovered.

        Parameters
        ----------

        n_components: int >= 1 or float > 0.0, optional
            The number of components that are kept or else the amount (ratio)
            of variance that is kept. If None, `self.n_active_components` is
            used.

        Notes
        -----
        In case `n_components` is greater than the total number of
        components or greater than the amount of variance
        currently kept, this method does not perform any action.
        """
        if n_components is None:
            # by default trim using the current n_active_components
            n_components = self.n_active_components
        # set self.n_active_components to n_components
        self.n_active_components = n_components

        if self.n_active_components < self.n_components:
            # set self.n_components to n_components
            self._components = self._components[:self.n_active_components]
            # store the eigenvalues associated to the discarded components
            self._trimmed_eigenvalues = \
                self._eigenvalues[self.n_active_components:]
            # make sure that the eigenvalues are trimmed too
            self._eigenvalues = self._eigenvalues[:self.n_active_components]

    def distance_to_subspace(self, instance):
        """
        Returns a version of `instance` where all the basis of the model
        have been projected out and which has been scaled by the inverse of
        the `noise_variance`

        Parameters
        ----------
        instance : :class:`menpo.base.Vectorizable`
            A novel instance.

        Returns
        -------
        scaled_projected_out : `self.instance_class`
            A copy of `instance`, with all basis of the model projected out
            and scaled by the inverse of the `noise_variance`.
        """
        vec_instance = self.distance_to_subspace_vector(instance.as_vector())
        return instance.from_vector(vec_instance)

    def distance_to_subspace_vector(self, vector_instance):
        """
        Returns a version of `instance` where all the basis of the model
        have been projected out and which has been scaled by the inverse of
        the `noise_variance`.

        Parameters
        ----------
        vector_instance : (n_features,) ndarray
            A novel vector.

        Returns
        -------
        scaled_projected_out: (n_features,) ndarray
            A copy of `vector_instance` with all basis of the model projected
            out and scaled by the inverse of the `noise_variance`.
        """
        return (self.inverse_noise_variance *
                self.project_out_vectors(vector_instance))

    def project_whitened(self, instance):
        """
        Returns a sheared (non-orthogonal) reconstruction of `instance`.

        Parameters
        ----------
        instance : :class:`menpo.base.Vectorizable`
            A novel instance.

        Returns
        -------
        sheared_reconstruction : `self.instance_class`
            A sheared (non-orthogonal) reconstruction of `instance`.
        """
        vector_instance = self.project_whitened_vector(instance.as_vector())
        return instance.from_vector(vector_instance)

    def project_whitened_vector(self, vector_instance):
        """
        Returns a sheared (non-orthogonal) reconstruction of
        `vector_instance`.

        Parameters
        ----------
        vector_instance : (n_features,) ndarray
            A novel vector.

        Returns
        -------
        sheared_reconstruction : (n_features,) ndarray
            A sheared (non-orthogonal) reconstruction of `vector_instance`
        """
        whitened_components = self.whitened_components
        weights = dgemm(alpha=1.0, a=vector_instance.T,
                        b=whitened_components.T, trans_a=True)
        return dgemm(alpha=1.0, a=weights.T, b=whitened_components.T,
                     trans_a=True, trans_b=True)

    def orthonormalize_against_inplace(self, linear_model):
        r"""
        Enforces that the union of this model's components and another are
        both mutually orthonormal.

        Note that the model passed in is guaranteed to not have it's number
        of available components changed. This model, however, may loose some
        dimensionality due to reaching a degenerate state.

        The removed components will always be trimmed from the end of
        components (i.e. the components which capture the least variance).
        If trimming is performed, `n_components` and
        `n_available_components` would be altered - see
        :meth:`trim_components` for details.

        Parameters
        -----------
        linear_model : :class:`LinearModel`
            A second linear model to orthonormalize this against.
        """
        # take the QR decomposition of the model components
        Q = (np.linalg.qr(np.hstack((linear_model._components.T,
                                     self._components.T)))[0]).T
        # the model passed to us went first, so all it's components will
        # survive. Pull them off, and update the other model.
        linear_model.components = Q[:linear_model.n_components, :]
        # it's possible that all of our components didn't survive due to
        # degeneracy. We need to trim our components down before replacing
        # them to ensure the number of components is consistent (otherwise
        # the components setter will complain at us)
        n_available_components = Q.shape[0] - linear_model.n_components
        if n_available_components < self.n_components:
            # oh dear, we've lost some components from the end of our model.
            if self.n_active_components < n_available_components:
                # save the current number of active components
                n_active_components = self.n_active_components
            else:
                # save the current number of available components
                n_active_components = n_available_components
            # call trim_components to update our state.
            self.trim_components(n_components=n_available_components)
            if n_active_components < n_available_components:
                # reset the number of active components
                self.n_active_components = n_active_components

        # now we can set our own components with the updated orthogonal ones
        self.components = Q[linear_model.n_components:, :]

    def __str__(self):
        str_out = 'PCA Model \n'
        str_out = str_out + \
            ' - centered:             {}\n' \
            ' - biased:               {}\n' \
            ' - # features:           {}\n' \
            ' - # active components:  {}\n'.format(
            self.centered, self.biased, self.n_features,
            self.n_active_components)
        str_out = str_out + \
            ' - kept variance:        {:.2}  {:.1%}\n' \
            ' - noise variance:       {:.2}  {:.1%}\n'.format(
            self.variance, self.variance_ratio,
            self.noise_variance, self.noise_variance_ratio)
        str_out = str_out + \
            ' - total # components:   {}\n' \
            ' - components shape:     {}\n'.format(
            self.n_components, self.components.shape)
        return str_out



from __future__ import division
import abc
import numpy as np

from menpo.transform import Scale, Translation, GeneralizedProcrustesAnalysis
from menpo.model.pca import PCAModel
from menpo.visualize import print_dynamic, progress_bar_str
from menpo.fitmultilevel.featurefunctions import compute_features

from .functions import mean_pointcloud


class DeformableModelBuilder(object):
    r"""
    Abstract class with a set of functions useful to build a Deformable Model.
    """
    __metaclass__ = abc.ABCMeta

    @classmethod
    def check_n_levels(cls, n_levels):
        r"""
        Checks the number of pyramid levels that must be int > 0.
        """
        if not isinstance(n_levels, int) or n_levels < 1:
            raise ValueError("n_levels must be int > 0")

    @classmethod
    def check_downscale(cls, downscale):
        r"""
        Checks the downscale factor of the pyramid that must be >= 1.
        """
        if downscale < 1:
            raise ValueError("downscale must be >= 1")

    @classmethod
    def check_normalization_diagonal(cls, normalization_diagonal):
        r"""
        Checks the diagonal length used to normalize the images' size that
        must be >= 20.
        """
        if normalization_diagonal is not None and normalization_diagonal < 20:
            raise ValueError("normalization_diagonal must be >= 20")

    @classmethod
    def check_boundary(cls, boundary):
        r"""
        Checks the boundary added around the reference shape that must be
        int >= 0.
        """
        if not isinstance(boundary, int) or boundary < 0:
            raise ValueError("boundary must be >= 0")

    @classmethod
    def check_max_components(cls, max_components, n_levels, var_name):
        r"""
        Checks the maximum number of components per level either of the shape
        or the appearance model. It must be None or int or float or a list of
        those containing 1 or {n_levels} elements.
        """
        str_error = ("{} must be None or an int > 0 or a 0 <= float <= 1 or "
                     "a list of those containing 1 or {} elements").format(
                         var_name, n_levels)
        if not isinstance(max_components, list):
            max_components_list = [max_components] * n_levels
        elif len(max_components) == 1:
            max_components_list = [max_components[0]] * n_levels
        elif len(max_components) == n_levels:
            max_components_list = max_components
        else:
            raise ValueError(str_error)
        for comp in max_components_list:
            if comp is not None:
                if not isinstance(comp, int):
                    if not isinstance(comp, float):
                        raise ValueError(str_error)
        return max_components_list

    @classmethod
    def check_feature_type(cls, feature_type, n_levels, pyramid_on_features):
        r"""
        Checks the feature type per level.
        If pyramid_on_features is False, it must be a string or a
        function/closure or a list of those containing 1 or {n_levels}
        elements.
        If pyramid_on_features is True, it must be a string or a
        function/closure or a list of 1 of those.

        Parameters
        ----------
        n_levels: int
            The number of pyramid levels.
        pyramid_on_features: boolean
            If True, the pyramid will be applied to the feature image, so
            the user needs to define a single feature_type.
            If False, the pyramid will be applied to the intensities image and
            features will be extracted at each level, so the user can define
            a feature_type per level.

        Returns
        -------
        feature_type_list: list
            A list of feature types.
            If pyramid_on_features is True, the list will have length 1.
            If pyramid_on_features is False, the list will have length
            {n_levels}.
        """
        if not pyramid_on_features:
            feature_type_str_error = ("feature_type must be a str or a "
                                      "function/closure or a list of "
                                      "those containing 1 or {} "
                                      "elements").format(n_levels)
            if not isinstance(feature_type, list):
                feature_type_list = [feature_type] * n_levels
            elif len(feature_type) == 1:
                feature_type_list = [feature_type[0]] * n_levels
            elif len(feature_type) == n_levels:
                feature_type_list = feature_type
            else:
                raise ValueError(feature_type_str_error)
        else:
            feature_type_str_error = ("pyramid_on_features is enabled so "
                                      "feature_type must be a str or a "
                                      "function/closure or a list "
                                      "containing 1 of those")
            if not isinstance(feature_type, list):
                feature_type_list = [feature_type]
            elif len(feature_type) == 1:
                feature_type_list = feature_type
            else:
                raise ValueError(feature_type_str_error)
        for ft in feature_type_list:
            if (ft is not None and not isinstance(ft, str)
                    and not hasattr(ft, '__call__')):
                        raise ValueError(feature_type_str_error)
        return feature_type_list

    @abc.abstractmethod
    def build(self, images, group=None, label='all'):
        r"""
        Builds a Multilevel Deformable Model.
        """
        pass

    @classmethod
    def _normalization_wrt_reference_shape(cls, images, group, label,
                                           normalization_diagonal,
                                           interpolator, verbose=False):
        r"""
        Function that normalizes the images sizes with respect to the reference
        shape (mean shape) scaling. This step is essential before building a
        deformable model.

        The normalization includes:
        1) Computation of the reference shape as the mean shape of the images'
           landmarks.
        2) Scaling of the reference shape using the normalization_diagonal.
        3) Rescaling of all the images so that their shape's scale is in
           correspondence with the reference shape's scale.

        Parameters
        ----------
        images: list of :class:`menpo.image.MaskedImage`
            The set of landmarked images from which to build the model.
        group : string
            The key of the landmark set that should be used. If None,
            and if there is only one set of landmarks, this set will be used.
        label: string
            The label of of the landmark manager that you wish to use. If no
            label is passed, the convex hull of all landmarks is used.
        normalization_diagonal: int
            During building an AAM, all images are rescaled to ensure that the
            scale of their landmarks matches the scale of the mean shape.

            If int, it ensures that the mean shape is scaled so that the
            diagonal of the bounding box containing it matches the
            normalization_diagonal value.
            If None, the mean shape is not rescaled.

            Note that, because the reference frame is computed from the mean
            landmarks, this kwarg also specifies the diagonal length of the
            reference frame (provided that features computation does not change
            the image size).
        interpolator: string
            The interpolator that should be used to perform the warps.
        verbose: bool, Optional
            Flag that controls information and progress printing.

            Default: False

        Returns
        -------
        reference_shape : :map:`PointCloud`
            The reference shape that was used to resize all training images to
            a consistent object size.
        normalized_images : :map:`MaskedImage` list
            A list with the normalized images.
        """
        # the reference_shape is the mean shape of the images' landmarks
        if verbose:
            print_dynamic('- Computing reference shape')
        shapes = [i.landmarks[group][label].lms for i in images]
        reference_shape = mean_pointcloud(shapes)

        # fix the reference_shape's diagonal length if asked
        if normalization_diagonal:
            x, y = reference_shape.range()
            scale = normalization_diagonal / np.sqrt(x**2 + y**2)
            Scale(scale, reference_shape.n_dims).apply_inplace(reference_shape)

        # normalize the scaling of all images wrt the reference_shape size
        normalized_images = []
        for c, i in enumerate(images):
            if verbose:
                print_dynamic('- Normalizing images size: {}'.format(
                    progress_bar_str((c + 1.) / len(images),
                                     show_bar=False)))
            normalized_images.append(i.rescale_to_reference_shape(
                reference_shape, group=group, label=label,
                interpolator=interpolator))

        if verbose:
            print_dynamic('- Normalizing images size: Done\n')
        return reference_shape, normalized_images

    @classmethod
    def _create_pyramid(cls, images, n_levels, downscale, pyramid_on_features,
                        feature_type, verbose=False):
        r"""
        Function that creates a generator function for Gaussian pyramid. The
        pyramid can be created either on the feature space or the original
        (intensities) space.

        Parameters
        ----------
        images: list of :class:`menpo.image.Image`
            The set of landmarked images from which to build the AAM.
        n_levels: int
            The number of multi-resolution pyramidal levels to be used.
        downscale: float
            The downscale factor that will be used to create the different
            pyramidal levels.
        pyramid_on_features: boolean
            If True, the features are extracted at the highest level and the
            pyramid is created on the feature images.
            If False, the pyramid is created on the original (intensities)
            space.
        feature_type: list of size 1 with str or function/closure or None
            The feature type to be used in case pyramid_on_features is enabled.
        verbose: bool, Optional
            Flag that controls information and progress printing.

            Default: False

        Returns
        -------
        generator: function
            The generator function of the Gaussian pyramid.
        """
        if pyramid_on_features:
            # compute features at highest level
            feature_images = []
            for c, i in enumerate(images):
                if verbose:
                    print_dynamic('- Computing feature space: {}'.format(
                        progress_bar_str((c + 1.) / len(images),
                                         show_bar=False)))
                feature_images.append(compute_features(i, feature_type[0]))
            if verbose:
                print_dynamic('- Computing feature space: Done\n')

            # create pyramid on feature_images
            generator = [i.gaussian_pyramid(n_levels=n_levels,
                                            downscale=downscale)
                         for i in feature_images]
        else:
            # create pyramid on intensities images
            # features will be computed per level
            generator = [i.gaussian_pyramid(n_levels=n_levels,
                                            downscale=downscale)
                         for i in images]
        return generator

    #TODO: this seems useful on its own, maybe it shouldn't be underscored...
    @classmethod
    def _build_shape_model(cls, shapes, max_components):
        r"""
        Builds a shape model given a set of shapes.

        Parameters
        ----------
        shapes: list of :map:`PointCloud`
            The set of shapes from which to build the model.
        max_components: None or int or float
            Specifies the number of components of the trained shape model.
            If int, it specifies the exact number of components to be retained.
            If float, it specifies the percentage of variance to be retained.
            If None, all the available components are kept (100% of variance).

        Returns
        -------
        shape_model: :class:`menpo.model.pca`
            The PCA shape model.
        """
        # centralize shapes
        centered_shapes = [Translation(-s.centre).apply(s) for s in shapes]
        # align centralized shape using Procrustes Analysis
        gpa = GeneralizedProcrustesAnalysis(centered_shapes)
        aligned_shapes = [s.aligned_source for s in gpa.transforms]

        # build shape model
        shape_model = PCAModel(aligned_shapes)
        if max_components is not None:
            # trim shape model if required
            shape_model.trim_components(max_components)

        return shape_model



from scipy.linalg import norm
import numpy as np

from .base import AppearanceLucasKanade


class AlternatingForwardAdditive(AppearanceLucasKanade):

    @property
    def algorithm(self):
        return 'Alternating-FA'

    def _fit(self, fitting_result, max_iters=20):
        # Initial error > eps
        error = self.eps + 1
        image = fitting_result.image
        fitting_result.weights = [[0]]
        n_iters = 0

        # Forward Additive Algorithm
        while n_iters < max_iters and error > self.eps:
            # Compute warped image with current weights
            IWxp = image.warp_to(self.template.mask, self.transform,
                                 interpolator=self.interpolator)

            # Compute appearance
            weights = self.appearance_model.project(IWxp)
            self.template = self.appearance_model.instance(weights)
            fitting_result.weights.append(weights)

            # Compute warp Jacobian
            dW_dp = self.transform.d_dp(self.template.mask.true_indices)

            # Compute steepest descent images, VI_dW_dp
            self._J = self.residual.steepest_descent_images(
                image, dW_dp, forward=(self.template, self.transform,
                                       self.interpolator))

            # Compute Hessian and inverse
            self._H = self.residual.calculate_hessian(self._J)

            # Compute steepest descent parameter updates
            sd_delta_p = self.residual.steepest_descent_update(
                self._J, self.template, IWxp)

            # Compute gradient descent parameter updates
            delta_p = np.real(self._calculate_delta_p(sd_delta_p))

            # Update warp weights
            parameters = self.transform.as_vector() + delta_p
            self.transform.from_vector_inplace(parameters)
            fitting_result.parameters.append(parameters)

            # Test convergence
            error = np.abs(norm(delta_p))
            n_iters += 1

        fitting_result.fitted = True
        return fitting_result


class AlternatingForwardCompositional(AppearanceLucasKanade):

    @property
    def algorithm(self):
        return 'Alternating-FC'

    def _set_up(self):
        # Compute warp Jacobian
        self._dW_dp = self.transform.d_dp(self.template.mask.true_indices)

    def _fit(self, fitting_result, max_iters=20):
        # Initial error > eps
        error = self.eps + 1
        image = fitting_result.image
        fitting_result.weights = [[0]]
        n_iters = 0

        # Forward Additive Algorithm
        while n_iters < max_iters and error > self.eps:
            # Compute warped image with current weights
            IWxp = image.warp_to(self.template.mask, self.transform,
                                 interpolator=self.interpolator)

            # Compute template by projection
            weights = self.appearance_model.project(IWxp)
            self.template = self.appearance_model.instance(weights)
            fitting_result.weights.append(weights)

            # Compute steepest descent images, VI_dW_dp
            self._J = self.residual.steepest_descent_images(IWxp, self._dW_dp)

            # Compute Hessian and inverse
            self._H = self.residual.calculate_hessian(self._J)

            # Compute steepest descent parameter updates
            sd_delta_p = self.residual.steepest_descent_update(
                self._J, self.template, IWxp)

            # Compute gradient descent parameter updates
            delta_p = np.real(self._calculate_delta_p(sd_delta_p))

            # Update warp weights
            self.transform.compose_after_from_vector_inplace(delta_p)
            fitting_result.parameters.append(self.transform.as_vector())

            # Test convergence
            error = np.abs(norm(delta_p))
            n_iters += 1

        fitting_result.fitted = True
        return fitting_result


class AlternatingInverseCompositional(AppearanceLucasKanade):

    @property
    def algorithm(self):
        return 'Alternating-IC'

    def _set_up(self):
        # Compute warp Jacobian
        self._dW_dp = self.transform.d_dp(self.template.mask.true_indices)

    def _fit(self, fitting_result, max_iters=20):
        # Initial error > eps
        error = self.eps + 1
        image = fitting_result.image
        fitting_result.weights = [[0]]
        n_iters = 0

        # Baker-Matthews, Inverse Compositional Algorithm
        while n_iters < max_iters and error > self.eps:
            # Compute warped image with current weights
            IWxp = image.warp_to(self.template.mask, self.transform,
                                 interpolator=self.interpolator)

            # Compute appearance
            weights = self.appearance_model.project(IWxp)
            self.template = self.appearance_model.instance(weights)
            fitting_result.weights.append(weights)

            # Compute steepest descent images, VT_dW_dp
            self._J = self.residual.steepest_descent_images(self.template,
                                                            self._dW_dp)

            # Compute Hessian and inverse
            self._H = self.residual.calculate_hessian(self._J)

            # Compute steepest descent parameter updates
            sd_delta_p = self.residual.steepest_descent_update(
                self._J, IWxp, self.template)

            # Compute gradient descent parameter updates
            delta_p = np.real(self._calculate_delta_p(sd_delta_p))

            # Request the pesudoinverse vector from the transform
            inv_delta_p = self.transform.pseudoinverse_vector(delta_p)

            # Update warp weights
            self.transform.compose_after_from_vector_inplace(inv_delta_p)
            fitting_result.parameters.append(self.transform.as_vector())

            # Test convergence
            error = np.abs(norm(delta_p))
            n_iters += 1

        fitting_result.fitted = True
        return fitting_result



import abc

from menpo.fit.base import Fitter
from menpo.fit.fittingresult import (NonParametricFittingResult,
                                     SemiParametricFittingResult,
                                     ParametricFittingResult)


class Regressor(Fitter):
    r"""
    An abstract base class for fitting Regressors.

    Parameters
    ----------
    regressor: function/closure
        The regressor to be used from
        `menpo.fit.regression.regressionfunctions.py`.
    features:
        The features used to regress.
    """
    def __init__(self, regressor, features):
        self.regressor = regressor
        self.features = features

    def _set_up(self):
        r"""
        Abstract method that sets up the fitter object.
        """
        pass

    def _fit(self, fitting_result, max_iters=1):
        r"""
        Abstract method to fit an image.

        Parameters
        ----------
        fitting_result: `menpo.fit.fittingresult`
            The fitting result object.
        max_iters: int
            The maximum number of iterations.
        """
        image = fitting_result.image
        initial_shape = fitting_result.initial_shape
        n_iters = 0

        while n_iters < max_iters:
            features = self.features(image, initial_shape)
            delta_p = self.regressor(features)

            fitted_shape, parameters = self.update(delta_p, initial_shape)
            fitting_result.parameters.append(parameters)
            n_iters += 1

        fitting_result.fitted = True
        return fitting_result

    @abc.abstractmethod
    def update(self, delta_p, initial_shape):
        r"""
        Abstract method to update the parameters.
        """
        pass


class NonParametricRegressor(Regressor):
    r"""
    Fitter of Non-Parametric Regressor.

    Parameters
    ----------
    regressor: function/closure
        The regressor to be used from
        `menpo.fit.regression.regressionfunctions.py`.
    features:
        The features used to regress.
    """
    def __init__(self, regressor, features):
        super(NonParametricRegressor, self).__init__(
            regressor, features)

    @property
    def algorithm(self):
        r"""
        Returns the regression type.
        """
        return "Non-Parametric"

    def _create_fitting_result(self, image, shapes, gt_shape=None):
        r"""
        Creates the fitting result object.

        Parameters
        ----------
        image: :class:`menpo.image.MaskedImage`
            The current image..
        shape: :class:`menpo.shape.PointCloud`
            The current shape.
        gt_shape: :class:`menpo.shape.PointCloud`
            The ground truth shape.
        """
        return NonParametricFittingResult(image, self, shapes=[shapes],
                                          gt_shape=gt_shape)

    def update(self, delta_shape, initial_shape):
        r"""
        Updates the shape.

        Parameters
        ----------
        delta_shape: PointCloud
            The shape increment.
        initial_shape: PointCloud
            The current shape.
        """
        fitted_shape = initial_shape.from_vector(
            initial_shape.as_vector() + delta_shape)
        return fitted_shape, fitted_shape

    def get_parameters(self, shape):
        r"""
        Method that makes sure that the parameter passed to the fit method is
        the shape.

        Parameters
        ----------
        shape: PointCloud
            The current shape.
        """
        return shape


class SemiParametricRegressor(Regressor):
    r"""
    Fitter of Semi-Parametric Regressor.

    Parameters
    ----------
    regressor: function/closure
        The regressor to be used from
        `menpo.fit.regression.regressionfunctions.py`.
    features:
        The features used to regress.
    """
    def __init__(self, regressor, features, transform, update='composition'):
        super(SemiParametricRegressor, self).__init__(
            regressor, features)
        self.transform = transform
        self._update = self._select_update(update)

    @property
    def algorithm(self):
        r"""
        Returns the regression type.
        """
        return "Semi-Parametric"

    def _create_fitting_result(self, image, parameters, gt_shape=None):
        r"""
        Creates the fitting result object.

        Parameters
        ----------
        image: :class:`menpo.image.MaskedImage`
            The current image..
        parameters: numpy.ndarray
            The current parameters vector.
        gt_shape: :class:`menpo.shape.PointCloud`, Optional
            The ground truth shape.

            Default: None
        """
        self.transform.from_vector_inplace(parameters)
        return SemiParametricFittingResult(
            image, self, parameters=[self.transform.as_vector()],
            gt_shape=gt_shape)

    def fit(self, image, initial_parameters, gt_shape=None, **kwargs):
        self.transform.from_vector_inplace(initial_parameters)
        return Fitter.fit(self, image, initial_parameters, gt_shape=gt_shape,
                          **kwargs)

    def _select_update(self, update):
        r"""
        Select the way to update the parameters.

        Parameters
        ----------
        update : {'compositional', 'additive'}
            The update method.

        Returns
        -------
        update : `function`
            The correct function to apply the update chosen.
        """
        if update == 'additive':
            return self._additive
        elif update == 'compositional':
            return self._compositional
        else:
            raise ValueError('Unknown update string selected. Valid'
                             'options are: additive, compositional')

    def _additive(self, delta_p):
        r"""
        Updates the parameters in the additive way.

        Parameters
        ----------
        delta_p : `ndarray`
            The parameters increment
        """
        parameters = self.transform.as_vector() + delta_p
        self.transform.from_vector_inplace(parameters)

    def _compositional(self, delta_p):
        r"""
        Updates the parameters in the compositional way.

        Parameters
        ----------
        delta_p : `ndarray`
            The parameters increment
        """
        self.transform.compose_after_from_vector_inplace(delta_p)

    def update(self, delta_p, initial_shape):
        r"""
        Updates the parameters of the shape model.

        Parameters
        ----------
        delta_p : `ndarray`
            The parameters increment.

        initial_shape : :map:`PointCloud`
            The current shape.
        """
        self._update(delta_p)
        return self.transform.target, self.transform.as_vector()

    def get_parameters(self, shape):
        r"""
        Method that makes sure that the parameter passed to the fit method is
        the model parameters.

        Parameters
        ----------
        shape : :map:`PointCloud`
            The current shape.
        """
        self.transform.set_target(shape)
        return self.transform.as_vector()


class ParametricRegressor(SemiParametricRegressor):
    r"""
    Fitter of Parametric Regressor.

    Parameters
    ----------
    regressor: function/closure
        The regressor to be used from
        `menpo.fit.regression.regressionfunctions.py`.
    features:
        The features used to regress.
    """
    def __init__(self, regressor, features, appearance_model, transform,
                 update='composition'):
        super(ParametricRegressor, self).__init__(
            regressor, features, transform, update=update)
        self.appearance_model = appearance_model
        self.template = appearance_model.mean

    @property
    def algorithm(self):
        r"""
        Returns the regression type.
        """
        return "Parametric"

    def _create_fitting_result(self, image, parameters, gt_shape=None):
        r"""
        Creates the fitting result object.

        Parameters
        ----------
        image: :class:`menpo.image.MaskedImage`
            The current image..
        parameters: numpy.ndarray
            The current parameters vector.
        gt_shape: :class:`menpo.shape.PointCloud`, Optional
            The ground truth shape.

            Default: None
        """
        self.transform.from_vector_inplace(parameters)
        return ParametricFittingResult(
            image, self, parameters=[self.transform.as_vector()],
            gt_shape=gt_shape)



import abc
import copy
import numpy as np

from menpo.base import DX, DP
from .base import Homogeneous, HomogFamilyAlignment


class Affine(Homogeneous, DP, DX):
    r"""Base class for all n-dimensional affine transformations. Provides
    methods to break the transform down into it's constituent
    scale/rotation/translation, to view the homogeneous matrix equivalent,
    and to chain this transform with other affine transformations.

    Parameters
    ----------
    h_matrix : ``(n_dims + 1, n_dims + 1)`` `ndarray`
        The homogeneous matrix of the affine transformation.

    copy : `bool`, optional
        If ``False`` avoid copying ``h_matrix`` for performance.

    skip_checks : `bool`, optional
        If ``True`` avoid sanity checks on ``h_matrix`` for performance.
    """
    def __init__(self, h_matrix, copy=True, skip_checks=False):
        Homogeneous.__init__(self, h_matrix, copy=copy,
                             skip_checks=skip_checks)

    @classmethod
    def identity(cls, n_dims):
        return Affine(np.eye(n_dims + 1))

    @property
    def h_matrix(self):
        return self._h_matrix

    def _set_h_matrix(self, value, copy=True, skip_checks=False):
        r"""Updates the h_matrix, performing sanity checks.

        Parameters
        ----------
        value : ndarray
            The new homogeneous matrix to set

        copy : `bool`, optional
            If False do not copy the h_matrix. Useful for performance.

        skip_checks : `bool`, optional
            If True skip sanity checks on the matrix. Useful for performance.
        """
        if not skip_checks:
            shape = value.shape
            if len(shape) != 2 or shape[0] != shape[1]:
                raise ValueError("You need to provide a square homogeneous "
                                 "matrix")
            if self.h_matrix is not None:
                # already have a matrix set! The update better be the same size
                if self.n_dims != shape[0] - 1:
                    raise ValueError("Trying to update the homogeneous "
                                     "matrix to a different dimension")
            if shape[0] - 1 not in [2, 3]:
                raise ValueError("Affine Transforms can only be 2D or 3D")
            if not (np.allclose(value[-1, :-1], 0) and
                    np.allclose(value[-1, -1], 1)):
                raise ValueError("Bottom row must be [0 0 0 1] or [0, 0, 1]")
        if copy:
            value = value.copy()
        self._h_matrix = value

    @property
    def linear_component(self):
        r"""The linear component of this affine transform.

        :type: ``(n_dims, n_dims)`` `ndarray`
        """
        return self.h_matrix[:-1, :-1]

    @property
    def translation_component(self):
        r"""The translation component of this affine transform.

        :type: ``(n_dims,)`` `ndarray`
        """
        return self.h_matrix[:-1, -1]

    def decompose(self):
        r"""Decompose this transform into discrete Affine Transforms.

        Useful for understanding the effect of a complex composite transform.

        Returns
        -------
        transforms : list of :map:`DiscreteAffine`
            Equivalent to this affine transform, such that::

                reduce(lambda x,y: x.chain(y), self.decompose()) == self
        """
        from .rotation import Rotation
        from .translation import Translation
        from .scale import Scale
        U, S, V = np.linalg.svd(self.linear_component)
        rotation_2 = Rotation(U)
        rotation_1 = Rotation(V)
        scale = Scale(S)
        translation = Translation(self.translation_component)
        return [rotation_1, scale, rotation_2, translation]

    def _transform_str(self):
        r"""
        A string representation explaining what this affine transform does.
        Has to be implemented by base classes.

        Returns
        -------
        str : string
            String representation of transform.
        """
        header = 'Affine decomposing into:'
        list_str = [t._transform_str() for t in self.decompose()]
        return header + reduce(lambda x, y: x + '\n' + '  ' + y, list_str, '  ')

    def _apply(self, x, **kwargs):
        r"""
        Applies this transform to a new set of vectors.

        Parameters
        ----------
        x : (N, D) ndarray
            Array to apply this transform to.


        Returns
        -------
        transformed_x : (N, D) ndarray
            The transformed array.
        """
        return np.dot(x, self.linear_component.T) + self.translation_component

    @property
    def n_parameters(self):
        r"""
        `n_dims * (n_dims + 1)` parameters - every element of the matrix bar
        the homogeneous part.

        :type: int

        Examples
        --------
        2D Affine: 6 parameters::

            [p1, p3, p5]
            [p2, p4, p6]

        3D Affine: 12 parameters::

            [p1, p4, p7, p10]
            [p2, p5, p8, p11]
            [p3, p6, p9, p12]
        """
        return self.n_dims * (self.n_dims + 1)

    def _as_vector(self):
        r"""
        Return the parameters of the transform as a 1D array. These parameters
        are parametrised as deltas from the identity warp. This does not
        include the homogeneous part of the warp. Note that it flattens using
        Fortran ordering, to stay consistent with Matlab.

        **2D**

        ========= ===========================================
        parameter definition
        ========= ===========================================
        p1        Affine parameter
        p2        Affine parameter
        p3        Affine parameter
        p4        Affine parameter
        p5        Translation in `x`
        p6        Translation in `y`
        ========= ===========================================

        3D and higher transformations follow a similar format to the 2D case.

        Returns
        -------
        params : ``(n_parameters,)`` `ndarray`
            The values that parametrise the transform.
        """
        params = self.h_matrix - np.eye(self.n_dims + 1)
        return params[:self.n_dims, :].ravel(order='F')

    def from_vector_inplace(self, p):
        r"""
        Updates this Affine in-place from the new parameters. See
        from_vector for details of the parameter format
        """
        h_matrix = None
        if p.shape[0] == 6:  # 2D affine
            h_matrix = np.eye(3)
            h_matrix[:2, :] += p.reshape((2, 3), order='F')
        elif p.shape[0] == 12:  # 3D affine
            h_matrix = np.eye(4)
            h_matrix[:3, :] += p.reshape((3, 4), order='F')
        else:
            ValueError("Only 2D (6 parameters) or 3D (12 parameters) "
                       "homogeneous matrices are supported.")
        self.set_h_matrix(h_matrix, copy=False, skip_checks=True)

    @property
    def composes_inplace_with(self):
        return Affine

    def _build_pseudoinverse(self):
        # Skip the checks as we know inverse of a homogeneous is a homogeneous
        return Affine(np.linalg.inv(self.h_matrix), copy=False,
                      skip_checks=True)

    def d_dp(self, points):
        r"""The first order derivative of this Affine transform wrt parameter
        changes evaluated at points.

        The Jacobian generated (for 2D) is of the form::

            x 0 y 0 1 0
            0 x 0 y 0 1

        This maintains a parameter order of::

          W(x;p) = [1 + p1  p3      p5] [x]
                   [p2      1 + p4  p6] [y]
                                        [1]

        Parameters
        ----------
        points : (n_points, n_dims) ndarray
            The set of points to calculate the jacobian for.


        Returns
        -------
        (n_points, n_params, n_dims) ndarray
            The jacobian wrt parametrization

        """
        n_points, points_n_dim = points.shape
        if points_n_dim != self.n_dims:
            raise ValueError(
                "Trying to sample jacobian in incorrect dimensions "
                "(transform is {0}D, sampling at {1}D)".format(
                    self.n_dims, points_n_dim))
        # prealloc the jacobian
        jac = np.zeros((n_points, self.n_parameters, self.n_dims))
        # a mask that we can apply at each iteration
        dim_mask = np.eye(self.n_dims, dtype=np.bool)

        for i, s in enumerate(
                range(0, self.n_dims * self.n_dims, self.n_dims)):
            # i is current axis
            # s is slicing offset
            # make a mask for a single points jacobian
            full_mask = np.zeros((self.n_parameters, self.n_dims), dtype=bool)
            # fill the mask in for the ith axis
            full_mask[slice(s, s + self.n_dims)] = dim_mask
            # assign the ith axis points to this mask, broadcasting over all
            # points
            jac[:, full_mask] = points[:, i][..., None]
            # finally, just repeat the same but for the ones at the end
        full_mask = np.zeros((self.n_parameters, self.n_dims), dtype=bool)
        full_mask[slice(s + self.n_dims, s + 2 * self.n_dims)] = dim_mask
        jac[:, full_mask] = 1
        return jac

    def d_dx(self, points):
        r"""
        The first order derivative of this Affine transform wrt spatial changes
        evaluated at points.

        The Jacobian for a given point (for 2D) is of the form::

            Jx = [(1 + a),     -b  ]
            Jy = [   b,     (1 + a)]
            J =  [Jx, Jy] = [[(1 + a), -b], [b, (1 + a)]]

        where a and b come from:

            W(x;p) = [1 + a   -b      tx] [x]
                     [b       1 + a   ty] [y]
                                          [1]
        Hence it is simply the linear component of the transform.

        Parameters
        ----------

        points: ndarray shape (n_points, n_dims)
            The spatial points at which the derivative should be evaluated.

        Returns
        -------

        d_dx: (1, n_dims, n_dims) ndarray
            The jacobian wrt spatial changes.

            d_dx[0, j, k] is the scalar differential change that the
            j'th dimension of the i'th point experiences due to a first order
            change in the k'th dimension.

            Note that because the jacobian is constant across space the first
            axis is length 1 to allow for broadcasting.

        """
        return self.linear_component[None, ...]


class AlignmentAffine(HomogFamilyAlignment, Affine):
    r"""
    Constructs an Affine by finding the optimal affine transform to align
    source to target.

    Parameters
    ----------
    source : :map:`PointCloud`
        The source pointcloud instance used in the alignment

    target : :map:`PointCloud`
        The target pointcloud instance used in the alignment

    Notes
    -----
    We want to find the optimal transform M which satisfies

        M a = b

    where `a` and `b` are the source and target homogeneous vectors
    respectively.

       (M a)' = b'
       a' M' = b'
       a a' M' = a b'

       `a a'` is of shape `(n_dim + 1, n_dim + 1)` and so can be inverted
       to solve for M.

       This approach is the analytical linear least squares solution to
       the problem at hand. It will have a solution as long as `(a a')`
       is non-singular, which generally means at least 2 corresponding
       points are required.
    """
    def __init__(self, source, target):
        # first, initialize the alignment
        HomogFamilyAlignment.__init__(self, source, target)
        # now, the Affine
        optimal_h = self._build_alignment_h_matrix(source, target)
        Affine.__init__(self, optimal_h, copy=False, skip_checks=True)

    @staticmethod
    def _build_alignment_h_matrix(source, target):
        r"""
        Returns the optimal alignment of source to target.
        """
        a = source.h_points
        b = target.h_points
        return np.linalg.solve(np.dot(a, a.T), np.dot(a, b.T)).T

    def set_h_matrix(self, value, copy=True, skip_checks=False):
        r"""
        Updates ``h_matrix``, optionally performing sanity checks.

        .. note::

            Updating the ``h_matrix`` on an :map:`AlignmentAffine`
            triggers a sync of the target.

        Note that it won't always be possible to manually specify the
        ``h_matrix`` through this method, specifically if changing the
        ``h_matrix`` could change the nature of the transform. See
        :attr:`h_matrix_is_mutable` for how you can discover if the
        ``h_matrix`` is allowed to be set for a given class.

        Parameters
        ----------
        value : ndarray
            The new homogeneous matrix to set
        copy : `bool`, optional
            If False do not copy the h_matrix. Useful for performance.
        skip_checks : `bool`, optional
            If True skip checking. Useful for performance.

        Raises
        ------
        NotImplementedError
            If :attr:`h_matrix_is_mutable` returns ``False``.


        Parameters
        ----------

        value : ndarray
            The new homogeneous matrix to set

        copy : bool, optional
            If False do not copy the h_matrix. Useful for performance.

        skip_checks : bool, optional
            If True skip verification for performance.
        """
        Affine.set_h_matrix(self, value, copy=copy, skip_checks=skip_checks)
        # now update the state
        self._sync_target_from_state()

    def _sync_state_from_target(self):
        optimal_h = self._build_alignment_h_matrix(self.source, self.target)
        # Use the pure Affine setter (so we don't get syncing)
        # We know the resulting affine is correct so skip the checks
        Affine.set_h_matrix(self, optimal_h, copy=False, skip_checks=True)

    def as_non_alignment(self):
        r"""Returns a copy of this affine without it's alignment nature.

        Returns
        -------
        transform : :map:`Affine`
            A version of this affine with the same transform behavior but
            without the alignment logic.
        """
        return Affine(self.h_matrix, skip_checks=True)


class DiscreteAffine(object):
    r"""
    A discrete Affine transform operation (such as a :meth:`Scale`,
    :class:`Translation` or :meth:`Rotation`). Has to be able to invertable.
    Make sure you inherit from :class:`DiscreteAffine` first,
    for optimal `decompose()` behavior.
    """

    __metaclass__ = abc.ABCMeta

    def decompose(self):
        r"""
        A :class:`DiscreteAffine` is already maximally decomposed -
        return a copy of self in a list.

        Returns
        -------
        transform : :class:`DiscreteAffine`
            Deep copy of `self`.
        """
        return [self.copy()]



import abc
from menpo.base import Vectorizable
from menpo.landmark import Landmarkable
from menpo.transform.base import Transformable
from menpo.visualize.base import Viewable


class Shape(Vectorizable, Landmarkable, Transformable, Viewable):
    """
    Abstract representation of shape. Shapes are :map:`Transformable`,
    :map:`Vectorizable`, :map:`Landmarkable` and :map:`Viewable`. This base
    class handles transforming landmarks when the shape is transformed.
    Therefore, implementations of :map:`Shape` have to implement the abstract
    :meth:`_transform_self_inplace` method that handles transforming the Shape
    itself.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(Shape, self).__init__()

    def _transform_inplace(self, transform):
        """
        Transform the landmarks and the shape itself.

        Parameters
        ----------
        transform : func
            A function to transform the spatial data with

        Returns
        -------
        self : `self`
            A pointer to `self` (the result of :meth:`_transform_self_inplace`).
        """
        self.landmarks._transform_inplace(transform)
        return self._transform_self_inplace(transform)

    @abc.abstractmethod
    def _transform_self_inplace(self, transform):
        """
        Implement this method to transform the concrete implementation of a
        shape. This is then called by the Shape's :meth:`_transform_inplace` method,
        which will have updated the landmarks beforehand.

        Parameters
        ----------
        transform : func
            A function to transform the spatial data with

        Returns
        -------
        self : `self`
            A pointer to `self`.
        """
        pass



from __future__ import division, print_function
import abc
import numpy as np

from menpo.image import Image
from menpo.fitmultilevel.functions import (noisy_align, build_sampling_grid,
                                           extract_local_patches_fast,
                                           extract_local_patches)
from menpo.fitmultilevel.featurefunctions import compute_features, sparse_hog
from menpo.fit.fittingresult import (NonParametricFittingResult,
                                     SemiParametricFittingResult,
                                     ParametricFittingResult)
from menpo.visualize import print_dynamic, progress_bar_str

from .base import (NonParametricRegressor, SemiParametricRegressor,
                   ParametricRegressor)
from .parametricfeatures import extract_parametric_features, weights
from .regressionfunctions import regression, mlr


class RegressorTrainer(object):
    r"""
    An abstract base class for training regressors.

    Parameters
    ----------
    reference_shape : :map:`PointCloud`
        The reference shape that will be used.

    regression_type : `function`, optional
        A `function` that defines the regression technique to be used.
        Examples of such closures can be found in
        :ref:`regression_functions`

    regression_features : ``None`` or `string` or `function`, optional
        The features that are used during the regression.

    noise_std : `float`, optional
        The standard deviation of the gaussian noise used to produce the
        training shapes.

    rotation : boolean, optional
        Specifies whether ground truth in-plane rotation is to be used
        to produce the training shapes.

    n_perturbations : `int`, optional
        Defines the number of perturbations that will be applied to the
        training shapes.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, reference_shape, regression_type=mlr,
                 regression_features=None, noise_std=0.04, rotation=False,
                 n_perturbations=10):
        self.reference_shape = reference_shape
        self.regression_type = regression_type
        self.regression_features = regression_features
        self.rotation = rotation
        self.noise_std = noise_std
        self.n_perturbations = n_perturbations

    def _regression_data(self, images, gt_shapes, perturbed_shapes,
                         verbose=False):
        r"""
        Method that generates the regression data : features and delta_ps.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images.

        gt_shapes : :map:`PointCloud` list
            List of the ground truth shapes that correspond to the images.

        perturbed_shapes : :map:`PointCloud` list
            List of the perturbed shapes in order to regress.

        verbose : `boolean`, optional
            If ``True``, the progress is printed.
        """
        if verbose:
            print_dynamic('- Generating regression data')

        n_images = len(images)
        features = []
        delta_ps = []
        for j, (i, s, p_shape) in enumerate(zip(images, gt_shapes,
                                                perturbed_shapes)):
            if verbose:
                print_dynamic('- Generating regression data - {}'.format(
                    progress_bar_str((j + 1.) / n_images, show_bar=False)))
            for ps in p_shape:
                features.append(self.features(i, ps))
                delta_ps.append(self.delta_ps(s, ps))
        return np.asarray(features), np.asarray(delta_ps)

    @abc.abstractmethod
    def features(self, image, shape):
        r"""
        Abstract method to generate the features for the regression.

        Parameters
        ----------
        image : :map:`MaskedImage`
            The current image.

        shape : :map:`PointCloud`
            The current shape.
        """
        pass

    @abc.abstractmethod
    def delta_ps(self, gt_shape, perturbed_shape):
        r"""
        Abstract method to generate the delta_ps for the regression.

        Parameters
        ----------
        gt_shape : :map:`PointCloud`
            The ground truth shape.

        perturbed_shape : :map:`PointCloud`
            The perturbed shape.
        """
        pass

    def train(self, images, shapes, perturbed_shapes=None, verbose=False,
              **kwargs):
        r"""
        Trains a Regressor given a list of landmarked images.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images from which to train the regressor.

        shapes : :map:`PointCloud` list
            List of the shapes that correspond to the images.

        perturbed_shapes : :map:`PointCloud` list, optional
            List of the perturbed shapes used for the regressor training.

        verbose : `boolean`, optional
            Flag that controls information and progress printing.

        Returns
        -------
        regressor : :map:`Regressor`
            A regressor object.

        Raises
        ------
        ValueError
            The number of shapes must be equal to the number of images.
        ValueError
            The number of perturbed shapes must be equal or multiple to
            the number of images.
        """
        n_images = len(images)
        n_shapes = len(shapes)

        # generate regression data
        if n_images != n_shapes:
            raise ValueError("The number of shapes must be equal to "
                             "the number of images.")
        elif not perturbed_shapes:
            perturbed_shapes = self.perturb_shapes(shapes)
            features, delta_ps = self._regression_data(
                images, shapes, perturbed_shapes, verbose=verbose)
        elif n_images == len(perturbed_shapes):
            features, delta_ps = self._regression_data(
                images, shapes, perturbed_shapes, verbose=verbose)
        else:
            raise ValueError("The number of perturbed shapes must be "
                             "equal or multiple to the number of images.")

         # perform regression
        if verbose:
            print_dynamic('- Performing regression...')
        regressor = regression(features, delta_ps, self.regression_type,
                               **kwargs)

        # compute regressor RMSE
        estimated_delta_ps = regressor(features)
        error = np.sqrt(np.mean(np.sum((delta_ps - estimated_delta_ps) ** 2,
                                       axis=1)))
        if verbose:
            print_dynamic('- Regression RMSE is {0:.5f}.\n'.format(error))
        return self._build_regressor(regressor, self.features)

    def perturb_shapes(self, gt_shape):
        r"""
        Perturbs the given shapes. The number of perturbations is defined by
        ``n_perturbations``.

        Parameters
        ----------
        gt_shape : :map:`PointCloud` list
            List of the shapes that correspond to the images.
            will be perturbed.

        Returns
        -------
        perturbed_shapes : :map:`PointCloud` list
            List of the perturbed shapes.
        """
        return [[self._perturb_shape(s) for _ in range(self.n_perturbations)]
                for s in gt_shape]

    def _perturb_shape(self, gt_shape):
        r"""
        Method that performs noisy alignment between the given ground truth
        shape and the reference shape.

        Parameters
        ----------
        gt_shape : :map:`PointCloud`
            The ground truth shape.
        """
        return noisy_align(self.reference_shape, gt_shape,
                           noise_std=self.noise_std
                           ).apply(self.reference_shape)

    @abc.abstractmethod
    def _build_regressor(self, regressor, features):
        r"""
        Abstract method to build a regressor model.
        """
        pass


class NonParametricRegressorTrainer(RegressorTrainer):
    r"""
    Class for training a Non-Parametric Regressor.

    Parameters
    ----------
    reference_shape : :map:`PointCloud`
        The reference shape that will be used.

    regression_type : `function`, optional
        A `function` that defines the regression technique to be used.
        Examples of such closures can be found in
        :ref:`regression_functions`

    regression_features : ``None`` or `string` or `function`, optional
        The features that are used during the regression.

        If ``None``, no feature representation will be computed from the
        original image.

        If `string` or `function`, the feature representation will be computed
        in the following way:
        
            If `string`, the feature representation will be extracted by
            executing::

                feature_image = getattr(image.features, feature_type)()

            For this to work properly feature_type needs to be one of
            Menpo's standard image feature methods. Note that, in this case,
            the feature computation will be carried out using its default
            options.

            Non-default feature options and new experimental feature can be
            used by defining a closure. In this case, the closure must define a
            function that receives as an input an image and returns a
            particular feature representation of that image. For example::

                def igo_double_from_std_normalized_intensities(image)
                    image = deepcopy(image)
                    image.normalize_std_inplace()
                    return image.feature_type.igo(double_angles=True)

            See :map:`ImageFeatures` for details more details on
            Menpo's standard image features and feature options.
            See :ref:`feature_functions` for non standard
            features definitions.

    patch_shape : tuple, optional
        The shape of the patches that will be extracted.

    noise_std : `float`, optional
        The standard deviation of the gaussian noise used to produce the
        training shapes.

    rotation : `boolean`, optional
        Specifies whether ground truth in-plane rotation is to be used
        to produce the training shapes.

    n_perturbations : `int`, optional
        Defines the number of perturbations that will be applied to the
        training shapes.

    """
    def __init__(self, reference_shape, regression_type=mlr,
                 regression_features=sparse_hog, patch_shape=(16, 16),
                 noise_std=0.04, rotation=False, n_perturbations=10):
        super(NonParametricRegressorTrainer, self).__init__(
            reference_shape, regression_type=regression_type,
            regression_features=regression_features, noise_std=noise_std,
            rotation=rotation, n_perturbations=n_perturbations)
        self.patch_shape = patch_shape
        self._set_up()

    def _set_up(self):
        # work out feature length per patch
        patch_img = Image.blank(self.patch_shape, fill=0)
        self._feature_patch_length = compute_features(
            patch_img, self.regression_features).n_parameters

    @property
    def algorithm(self):
        r"""
        Returns the algorithm name.
        """
        return "Non-Parametric"

    def _create_fitting(self, image, shapes, gt_shape=None):
        r"""
        Method that creates the fitting result object.

        Parameters
        ----------
        image : :map:`MaskedImage`
            The image object.

        shapes : :map:`PointCloud` list
            The shapes.

        gt_shape : :map:`PointCloud`
            The ground truth shape.
        """
        return NonParametricFittingResult(image, self, shapes=[shapes],
                                          gt_shape=gt_shape)

    def features(self, image, shape):
        r"""
        Method that extracts the features for the regression, which in this
        case are patch based.

        Parameters
        ----------
        image : :map:`MaskedImage`
            The current image.

        shape : :map:`PointCloud`
            The current shape.
        """
        # extract patches
        patches = extract_local_patches_fast(image, shape, self.patch_shape)

        features = np.zeros((shape.n_points, self._feature_patch_length))
        for j, patch in enumerate(patches):
            # build patch image
            patch_img = Image(patch, copy=False)
            # compute features
            features[j, ...] = compute_features(
                patch_img, self.regression_features).as_vector()

        return np.hstack((features.ravel(), 1))

    def delta_ps(self, gt_shape, perturbed_shape):
        r"""
        Method to generate the delta_ps for the regression.

        Parameters
        ----------
        gt_shape : :map:`PointCloud`
            The ground truth shape.

        perturbed_shape : :map:`PointCloud`
            The perturbed shape.
        """
        return (gt_shape.as_vector() -
                perturbed_shape.as_vector())

    def _build_regressor(self, regressor, features):
        r"""
        Method to build the NonParametricRegressor regressor object.
        """
        return NonParametricRegressor(regressor, features)


class SemiParametricRegressorTrainer(NonParametricRegressorTrainer):
    r"""
    Class for training a Semi-Parametric Regressor.

    This means that a parametric shape model and a non-parametric appearance
    representation are employed.

    Parameters
    ----------
    reference_shape : PointCloud
        The reference shape that will be used.

    regression_type : `function`, optional
        A `function` that defines the regression technique to be used.
        Examples of such closures can be found in
        :ref:`regression_functions`

    regression_features : ``None`` or `string` or `function`, optional
        The features that are used during the regression.

        If ``None``, no feature representation will be computed from the
        original image.

        If `string` or `function`, the feature representation will be computed
        in the following way:

            If `string`, the feature representation will be extracted by
            executing::

                feature_image = getattr(image.features, feature_type)()

            For this to work properly feature_type needs to be one of
            Menpo's standard image feature methods. Note that, in this case,
            the feature computation will be carried out using its default
            options.

            Non-default feature options and new experimental feature can be
            used by defining a closure. In this case, the closure must define a
            function that receives as an input an image and returns a
            particular feature representation of that image. For example::

                def igo_double_from_std_normalized_intensities(image)
                    image = deepcopy(image)
                    image.normalize_std_inplace()
                    return image.feature_type.igo(double_angles=True)

            See :map:`ImageFeatures` for details more details on
            Menpo's standard image features and feature options.
            See :ref:`feature_functions` for non standard
            features definitions.

    patch_shape : tuple, optional
        The shape of the patches that will be extracted.

    update : 'compositional' or 'additive'
        Defines the way to update the warp.

    noise_std : `float`, optional
        The standard deviation of the gaussian noise used to produce the
        training shapes.

    rotation : `boolean`, optional
        Specifies whether ground truth in-plane rotation is to be used
        to produce the training shapes.

    n_perturbations : `int`, optional
        Defines the number of perturbations that will be applied to the
        training shapes.

    """
    def __init__(self, transform, reference_shape, regression_type=mlr,
                 regression_features=sparse_hog, patch_shape=(16, 16),
                 update='compositional', noise_std=0.04, rotation=False,
                 n_perturbations=10):
        super(SemiParametricRegressorTrainer, self).__init__(
            reference_shape, regression_type=regression_type,
            regression_features=regression_features, patch_shape=patch_shape,
            noise_std=noise_std, rotation=rotation,
            n_perturbations=n_perturbations)
        self.transform = transform
        self.update = update

    @property
    def algorithm(self):
        r"""
        Returns the algorithm name.
        """
        return "Semi-Parametric"

    def _create_fitting(self, image, shapes, gt_shape=None):
        r"""
        Method that creates the fitting result object.

        Parameters
        ----------
        image : :map:`MaskedImage`
            The image object.

        shapes : :map:`PointCloud` list
            The shapes.

        gt_shape : :map:`PointCloud`
            The ground truth shape.
        """
        return SemiParametricFittingResult(image, self, parameters=[shapes],
                                           gt_shape=gt_shape)

    def delta_ps(self, gt_shape, perturbed_shape):
        r"""
        Method to generate the delta_ps for the regression.

        Parameters
        ----------
        gt_shape : :map:`PointCloud`
            The ground truth shape.

        perturbed_shape : :map:`PointCloud`
            The perturbed shape.
        """
        self.transform.set_target(gt_shape)
        gt_ps = self.transform.as_vector()
        self.transform.set_target(perturbed_shape)
        perturbed_ps = self.transform.as_vector()
        return gt_ps - perturbed_ps

    def _build_regressor(self, regressor, features):
        r"""
        Method to build the NonParametricRegressor regressor object.
        """
        return SemiParametricRegressor(regressor, features, self.transform,
                                       self.update)


class ParametricRegressorTrainer(RegressorTrainer):
    r"""
    Class for training a Parametric Regressor.

    Parameters
    ----------
    appearance_model : :map:`PCAModel`
        The appearance model to be used.

    transform : :map:`Affine`
        The transform used for warping.

    reference_shape : :map:`PointCloud`
        The reference shape that will be used.

    regression_type : `function`, optional
        A `function` that defines the regression technique to be used.
        Examples of such closures can be found in
        :ref:`regression_functions`

    regression_features : ``None`` or `function`, optional
        The parametric features that are used during the regression.

        If ``None``, the reconstruction appearance weights will be used as
        feature.

        If `string` or `function`, the feature representation will be
        computed using one of the function in:

            If `string`, the feature representation will be extracted by
            executing a parametric feature function.

            Note that this feature type can only be one of the parametric
            feature functions defined :ref:`parametric_features`.

    patch_shape : tuple, optional
        The shape of the patches that will be extracted.

    update : 'compositional' or 'additive'
        Defines the way to update the warp.

    noise_std : `float`, optional
        The standard deviation of the gaussian noise used to produce the
        training shapes.

    rotation : `boolean`, optional
        Specifies whether ground truth in-plane rotation is to be used
        to produce the training shapes.

    n_perturbations : `int`, optional
        Defines the number of perturbations that will be applied to the
        training shapes.

    interpolator : `string`
        Specifies the interpolator used in warping.

    """
    def __init__(self, appearance_model, transform, reference_shape,
                 regression_type=mlr, regression_features=weights,
                 update='compositional', noise_std=0.04, rotation=False,
                 n_perturbations=10, interpolator='scipy'):
        super(ParametricRegressorTrainer, self).__init__(
            reference_shape, regression_type=regression_type,
            regression_features=regression_features, noise_std=noise_std,
            rotation=rotation, n_perturbations=n_perturbations)
        self.appearance_model = appearance_model
        self.template = appearance_model.mean
        self.regression_features = regression_features
        self.transform = transform
        self.update = update
        self.interpolator = interpolator
    
    @property
    def algorithm(self):
        r"""
        Returns the algorithm name.
        """
        return "Parametric"

    def _create_fitting(self, image, shapes, gt_shape=None):
        r"""
        Method that creates the fitting result object.

        Parameters
        ----------
        image : :map:`MaskedImage`
            The image object.

        shapes : :map:`PointCloud` list
            The shapes.

        gt_shape : :map:`PointCloud`
            The ground truth shape.
        """
        return ParametricFittingResult(image, self, parameters=[shapes],
                                       gt_shape=gt_shape)

    def features(self, image, shape):
        r"""
        Method that extracts the features for the regression, which in this
        case are patch based.

        Parameters
        ----------
        image : :map:`MaskedImage`
            The current image.

        shape : :map:`PointCloud`
            The current shape.
        """
        self.transform.set_target(shape)
        warped_image = image.warp_to(self.template.mask, self.transform,
                                     interpolator=self.interpolator)
        features = extract_parametric_features(
            self.appearance_model, warped_image, self.regression_features)
        return np.hstack((features, 1))

    def delta_ps(self, gt_shape, perturbed_shape):
        r"""
        Method to generate the delta_ps for the regression.

        Parameters
        ----------
        gt_shape : :map:`PointCloud`
            The ground truth shape.

        perturbed_shape : :map:`PointCloud`
            The perturbed shape.
        """
        self.transform.set_target(gt_shape)
        gt_ps = self.transform.as_vector()
        self.transform.set_target(perturbed_shape)
        perturbed_ps = self.transform.as_vector()
        return gt_ps - perturbed_ps

    def _build_regressor(self, regressor, features):
        r"""
        Method to build the NonParametricRegressor regressor object.
        """
        return ParametricRegressor(
            regressor, features, self.appearance_model, self.transform,
            self.update)


class SemiParametricClassifierBasedRegressorTrainer(
        SemiParametricRegressorTrainer):
    r"""
    Class for training a Semi-Parametric Classifier-Based Regressor. This means
    that the classifiers are used instead of features.

    Parameters
    ----------
    classifiers : list of :ref:`classifier_functions`
        List of classifiers.

    transform : :map:`Affine`
        The transform used for warping.

    reference_shape : :map:`PointCloud`
        The reference shape that will be used.

    regression_type : `function`, optional
        A `function` that defines the regression technique to be used.
        Examples of such closures can be found in
        :ref:`regression_functions`

    patch_shape : tuple, optional
        The shape of the patches that will be extracted.

    noise_std : `float`, optional
        The standard deviation of the gaussian noise used to produce the
        training shapes.

    rotation : `boolean`, optional
        Specifies whether ground truth in-plane rotation is to be used
        to produce the training shapes.

    n_perturbations : `int`, optional
        Defines the number of perturbations that will be applied to the
        training shapes.

    interpolator : `string`
        Specifies the interpolator used in warping.
    """
    def __init__(self, classifiers, transform, reference_shape,
                 regression_type=mlr, patch_shape=(16, 16),
                 update='compositional', noise_std=0.04, rotation=False,
                 n_perturbations=10):
        super(SemiParametricClassifierBasedRegressorTrainer, self).__init__(
            transform, reference_shape, regression_type=regression_type,
            patch_shape=patch_shape, update=update,
            noise_std=noise_std,  rotation=rotation,
            n_perturbations=n_perturbations)
        self.classifiers = classifiers

    def _set_up(self):
        # TODO: CLMs should use slices instead of sampling grid, and the
        # need of the _set_up method will probably disappear
        # set up sampling grid
        self.sampling_grid = build_sampling_grid(self.patch_shape)

    def features(self, image, shape):
        r"""
        Method that extracts the features for the regression, which in this
        case are patch based.

        Parameters
        ----------
        image : :map:`MaskedImage`
            The current image.

        shape : :map:`PointCloud`
            The current shape.
        """
        # TODO: in the future this should be extract_local_patches_fast
        patches = extract_local_patches(image, shape, self.sampling_grid)
        features = [clf(np.reshape(patch, (-1, patch.shape[-1])))
                    for (clf, patch) in zip(self.classifiers, patches)]
        return np.hstack((np.asarray(features).ravel(), 1))


from __future__ import division
import numpy as np

from menpo.transform import Scale
from menpo.visualize.base import GraphPlotter, MultipleImageViewer
from menpo.fit.fittingresult import FittingResult


class MultilevelFittingResult(FittingResult):
    r"""
    Object that holds the state of a MultipleFitter object (to which it is
    linked) after it has fitted a particular image.

    Parameters
    -----------
    image: :class:`menpo.image.masked.MaskedImage`
        The fitted image.

    multiple_fitter: :class:`menpo.fitter.base.Fitter`
        The fitter object used to fitter the image.

    fitting_results: :class:`menpo.fitter.fittingresult.FittingResult` list
        A list of basic fitting objects.

    affine_correction: :class: `menpo.transforms.affine.Affine`
        An affine transform that maps the result of the top resolution
        fitting level to the space scale of the original image.

    gt_shape: class:`menpo.shape.PointCloud`, optional
        The ground truth shape associated to the image.

        Default: None

    error_type: 'me_norm', 'me' or 'rmse', optional.
        Specifies the way in which the error between the fitted and
        ground truth shapes is to be computed.

        Default: 'me_norm'
    """

    def __init__(self, image, multiple_fitter, fitting_results, affine_correction,
                 gt_shape=None, error_type='me_norm'):
        self._error_stop = None  # Internal attribute of error_type setter
        self.fitting_results = fitting_results
        self._affine_correction = affine_correction
        super(MultilevelFittingResult, self).__init__(
            image, multiple_fitter, gt_shape=gt_shape, error_type=error_type)

    @property
    def n_levels(self):
        r"""
        Returns the number of levels of the fitter object.
        """
        return self.fitter.n_levels

    @property
    def downscale(self):
        r"""
        Returns the downscale factor used by the multiple fitter.
        """
        return self.fitter.downscale

    @property
    def scaled_levels(self):
        r"""
        Returns True if the shape results returned by the basic fitting_results
        must be scaled.
        """
        return True  # self.fitter.scaled_levels

    @property
    def fitted(self):
        r"""
        Returns the fitted state of each fitting object.
        """
        return [f.fitted for f in self.fitting_results]

    @FittingResult.error_type.setter
    def error_type(self, error_type):
        r"""
        Sets the error type according to a set of predefined options.
        """
        if error_type == 'me_norm':
            for f in self.fitting_results:
                f.error_type = error_type
            self._error_stop = 0.1
            self._error_text = 'Point-to-point error normalized by object ' \
                               'size'
        elif error_type == 'me':
            NotImplementedError("erro_type 'me' not implemented yet")
        elif error_type == 'rmse':
            NotImplementedError("error_type 'rmse' not implemented yet")
        else:
            raise ValueError("Unknown error_type string selected. Valid"
                             "options are: 'me_norm', 'me', 'rmse'")
        self._error_type = error_type

    @property
    def n_iters(self):
        r"""
        Returns the total number of iterations used to fitter the image.
        """
        n_iters = 0
        for f in self.fitting_results:
            n_iters += f.n_iters
        return n_iters

    def shapes(self, as_points=False):
        r"""
        Generates a list containing the shapes obtained at each fitting
        iteration.

        Parameters
        -----------
        as_points: boolean, optional
            Whether the results is returned as a list of PointClouds or
            ndarrays.

            Default: False

        Returns
        -------
        shapes: :class:`menpo.shape.PointCoulds or ndarray list
            A list containing the shapes obtained at each fitting iteration.
        """
        n = self.n_levels - 1

        shapes = []
        for j, f in enumerate(self.fitting_results):
            if self.scaled_levels:
                transform = Scale(self.downscale**(n-j), 2)
                for t in f.shapes(as_points=as_points):
                    transform.apply_inplace(t)
                    shapes.append(self._affine_correction.apply(t))
            else:
                for t in f.shapes(as_points=as_points):
                    shapes.append(self._affine_correction.apply(t))

        return shapes

    @property
    def final_shape(self):
        r"""
        Returns the final fitted shape.
        """
        return self._affine_correction.apply(
            self.fitting_results[-1].final_shape)

    @property
    def initial_shape(self):
        r"""
        Returns the initial shape from which the fitting started.
        """
        n = self.n_levels - 1

        initial_shape = self.fitting_results[0].initial_shape
        if self.scaled_levels:
            Scale(self.downscale ** n,
                  initial_shape.n_dims).apply_inplace(initial_shape)

        return self._affine_correction.apply(initial_shape)

    @FittingResult.gt_shape.setter
    def gt_shape(self, value):
        r"""
        Setter for the ground truth shape associated to the image.
        """
        self._gt_shape = value

    def plot_error(self, figure_id=None, new_figure=False, **kwargs):
        r"""
        Plots the error evolution throughout the fitting.
        """
        if self.gt_shape is not None:
            title = 'Error evolution'
            legend = [self.algorithm]
            x_label = 'Number of iterations'
            y_label = self._error_text
            errors = self.errors
            x_limit = self.n_iters + self.n_levels
            axis_limits = [0, x_limit, 0, np.max(errors)]
            return GraphPlotter(figure_id, new_figure, range(0, x_limit),
                                [errors], title=title, legend=legend,
                                x_label=x_label, y_label=y_label,
                                axis_limits=axis_limits).render(**kwargs)
        else:
            raise ValueError('Ground truth shape has not been set, error '
                             'cannot be plotted')

    def __str__(self):
        out = "Initial error: {0:.4f}\nFinal error: {1:.4f}".format(
            self.initial_error, self.final_error)
        return out


class AAMMultilevelFittingResult(MultilevelFittingResult):
    r"""
    Object let us recover the state of an AAM Fitter after the latter has
    fitted a particular image.

    Parameters
    -----------
    image: :class:`pybug.image.masked.MaskedImage`
        The fitted image.

    aam_fitter: :class:`pybug.aam.fitter.AAMFitter`
        The aam_fitter object used to fitter the image.

    basic_fittings: :class:`pybug.aam.fitting.BasicFitting` list
        A list of basic fitting objects.

    _affine_correction: :class: `pybug.transforms.affine.Affine`
            An affine transform that maps the result of the top resolution
            fitting level to the space scale of the original image.

    gt_shape: class:`pybug.shape.PointCloud`, optional
        The ground truth shape associated to the image.

        Default: None

    error_type: 'me_norm', 'me' or 'rmse', optional.
        Specifies the way in which the error between the fitted and
        ground truth shapes is to be computed.

        Default: 'me_norm'
    """

    @property
    def residual(self):
        r"""
        Returns the type of residual used by the basic fitter associated to
        each basic fitting.
        """
        # TODO: ensure that all basic_fitting residuals are the same?
        return self.fitting_results[-1].residual.type

    @property
    def costs(self):
        r"""
        Returns a list containing the cost at each fitting iteration.
        """
        raise ValueError('costs not implemented yet.')
        #return self._flatten_out([f.costs for f in self.basic_fittings])

    @staticmethod
    def _flatten_out(list_of_lists):
        return [i for l in list_of_lists for i in l]

    @property
    def final_cost(self):
        r"""
        Returns the final fitting cost.
        """
        return self.fitting_results[-1].final_cost

    @property
    def initial_cost(self):
        r"""
        Returns the initial fitting cost.
        """
        return self.fitting_results[0].initial_cost

    def warped_images(self, from_basic_fittings=False, as_pixels=False):
        r"""
        Generates a list containing the warped images obtained at each fitting
        iteration.

        Parameters
        -----------
        from_basic_fittings : `boolean`, optional
            If ``True``, the returned transform per iteration is used to warp
            the internal image representation used by each basic fitter.
            If ``False``, the transforms are used to warp original image.

        as_pixels : `boolean`, optional
            Whether the result is returned as a list of :map:`Image` or
            `ndarray`.

        Returns
        -------
        warped_images : :map:`MaskedImage` or `ndarray` list
            A list containing the warped images obtained at each fitting
            iteration.
        """
        if from_basic_fittings:
            return self._flatten_out([f.warped_images(as_pixels=as_pixels)
                                      for f in self.fitting_results])
        else:
            mask = self.fitting_results[-1].fitter.template.mask
            transform = self.fitting_results[-1].fitter.transform
            interpolator = self.fitting_results[-1].fitter.interpolator
            warped_images = []
            for t in self.shapes():
                transform.set_target(t)
                image = self.image.warp_to(mask, transform,
                                           interpolator=interpolator)
                if as_pixels:
                    image = image.pixels

                warped_images.append(image)

        return warped_images

    def appearance_reconstructions(self, as_pixels=False):
        r"""
        Generates a list containing the appearance reconstruction obtained at
        each fitting iteration.

        Parameters
        ----------
        as_pixels : `boolean`, optional
            Whether the result is returned as a list of :map:`Image` or
            `ndarray`.

        Returns
        -------
        appearance_reconstructions : :map:`MaskedImage` or `ndarray` list
            A list containing the appearance reconstructions obtained at each
            fitting iteration.
        """
        return self._flatten_out(
            [f.appearance_reconstructions(as_pixels=as_pixels)
             for f in self.fitting_results])

    def plot_cost(self, figure_id=None, new_figure=False, **kwargs):
        r"""
        Plots the cost evolution throughout the fitting.
        """
        title = 'Cost evolution'
        legend = self.algorithm
        x_label = 'Number of iterations'
        y_label = 'Normalized cost'
        costs = [c for cost in self.costs for c in cost]
        total_n_iters = self.n_iters + self.n_levels
        axis_limits = [0, total_n_iters, 0, max(costs)]
        return GraphPlotter(figure_id, new_figure,
                            range(0, self.n_iters+self.n_levels), costs,
                            title=title, legend=legend, x_label=x_label,
                            y_label=y_label,
                            axis_limits=axis_limits).render(**kwargs)

    def view_warped_images(self, figure_id=None, new_figure=False,
                           channels=None, from_basic_fittings=False,
                           **kwargs):
        r"""
        Displays the warped images.
        """
        pixels_list = self.warped_images(
            from_basic_fittings=from_basic_fittings, as_pixels=True)
        return self._view_images(pixels_list, figure_id=figure_id,
                                 new_figure=new_figure, channels=channels,
                                 **kwargs)

    def view_appearance_reconstructions(self, figure_id=None,
                                        new_figure=False, channels=None,
                                        **kwargs):
        r"""
        Displays the appearance reconstructions.
        """
        pixels_list = self.appearance_reconstructions(as_pixels=True)
        return self._view_images(pixels_list, figure_id=figure_id,
                                 new_figure=new_figure, channels=channels,
                                 **kwargs)

    def view_error_images(self, figure_id=None, new_figure=False,
                          channels=None, **kwargs):
        r"""
        Displays the error images.
        """
        warped_images = self.warped_images(as_pixels=True)
        appearances = self.appearance_reconstructions(as_pixels=True)
        pixels_list = [a - i for a, i in zip(appearances, warped_images)]
        return self._view_images(pixels_list, figure_id=figure_id,
                                 new_figure=new_figure, channels=channels,
                                 **kwargs)

    def _view_images(self, pixels_list, figure_id=None, new_figure=False,
                     channels=None, **kwargs):
        return MultipleImageViewer(figure_id, new_figure,
                                   self.image.n_dims, pixels_list,
                                   channels=channels).render(**kwargs)

    def print_fitting_info(self):
        super(AAMMultilevelFittingResult, self).print_fitting_info()
        #print "Initial cost: {}".format(self.initial_cost)
        #print "Final cost: {}".format(self.final_cost)
